# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import loads

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import requests

from apprise import Apprise, NotifyType
from apprise.plugins.bulksms import NotifyBulkSMS

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    (
        "bulksms://",
        {
            # Instantiated but no auth, so no otification can happen
            "instance": NotifyBulkSMS,
            # Expected notify() response because we have no one to notify
            "notify_response": False,
        },
    ),
    (
        "bulksms://:@/",
        {
            # invalid auth
            "instance": NotifyBulkSMS,
            # Expected notify() response because we have no one to notify
            "notify_response": False,
        },
    ),
    (
        "bulksms://{}@12345678".format("a" * 10),
        {
            # Just user provided (no password)
            "instance": NotifyBulkSMS,
            # Expected notify() response because we have no one to notify
            "notify_response": False,
        },
    ),
    (
        "bulksms://{}:{}@{}".format("a" * 10, "b" * 10, "3" * 5),
        {
            # invalid nubmer provided
            "instance": NotifyBulkSMS,
            # Expected notify() response because we have no one to notify
            "notify_response": False,
        },
    ),
    (
        "bulksms://{}:{}@123/{}/abcd/".format("a" * 5, "b" * 10, "3" * 11),
        {
            # included group and phone, short number (123) dropped
            "instance": NotifyBulkSMS,
            "privacy_url": "bulksms://a...a:****@+33333333333/@abcd",
        },
    ),
    (
        "bulksms://{}:{}@{}?batch=y&unicode=n".format(
            "b" * 5, "c" * 10, "4" * 11
        ),
        {
            "instance": NotifyBulkSMS,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "bulksms://b...b:****@+4444444444",
        },
    ),
    (
        "bulksms://{}:{}@123456/{}".format("a" * 10, "b" * 10, "4" * 11),
        {
            # using short-code (6 characters)
            "instance": NotifyBulkSMS,
        },
    ),
    (
        "bulksms://{}:{}@{}".format("a" * 10, "b" * 10, "5" * 11),
        {
            # using phone no with no target - we text ourselves in
            # this case
            "instance": NotifyBulkSMS,
        },
    ),
    # Test route group
    (
        "bulksms://{}:{}@admin?route=premium".format("a" * 10, "b" * 10),
        {
            "instance": NotifyBulkSMS,
        },
    ),
    (
        "bulksms://{}:{}@admin?route=invalid".format("a" * 10, "b" * 10),
        {
            # invalid route
            "instance": TypeError,
        },
    ),
    (
        "bulksms://_?user={}&password={}&from={}".format(
            "a" * 10, "b" * 10, "5" * 11
        ),
        {
            # use get args to acomplish the same thing
            "instance": NotifyBulkSMS,
        },
    ),
    (
        "bulksms://_?user={}&password={}&from={}".format(
            "a" * 10, "b" * 10, "5" * 3
        ),
        {
            # use get args to acomplish the same thing
            "instance": TypeError,
        },
    ),
    (
        "bulksms://_?user={}&password={}&from={}&to={}".format(
            "a" * 10, "b" * 10, "5" * 11, "7" * 13
        ),
        {
            # use to=
            "instance": NotifyBulkSMS,
        },
    ),
    (
        "bulksms://{}:{}@{}".format("a" * 10, "b" * 10, "a" * 3),
        {
            "instance": NotifyBulkSMS,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "bulksms://{}:{}@{}".format("a" * 10, "b" * 10, "6" * 11),
        {
            "instance": NotifyBulkSMS,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_bulksms_urls():
    """NotifyTemplate() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_bulksms_edge_cases(mock_post):
    """NotifyBulkSMS() Edge Cases."""

    # Initialize some generic (but valid) tokens
    user = "abcd"
    pwd = "mypass123"
    targets = [
        "+1(555) 123-1234",
        "1555 5555555",
        "group",
        # A garbage entry
        "12",
        # NOw a valid one because a group was implicit
        "@12",
    ]

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_post.return_value = response

    # Instantiate our object
    obj = Apprise.instantiate(
        "bulksms://{}:{}@{}?batch=n".format(user, pwd, "/".join(targets))
    )

    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    # We know there are 4 targets
    assert len(obj) == 4

    # Test our call count
    assert mock_post.call_count == 4

    # Test
    details = mock_post.call_args_list[0]
    payload = loads(details[1]["data"])
    assert payload["to"] == "+15551231234"
    assert payload["body"] == "title\r\nbody"

    details = mock_post.call_args_list[1]
    payload = loads(details[1]["data"])
    assert payload["to"] == "+15555555555"
    assert payload["body"] == "title\r\nbody"

    details = mock_post.call_args_list[2]
    payload = loads(details[1]["data"])
    assert isinstance(payload["to"], dict)
    assert payload["to"]["name"] == "group"
    assert payload["body"] == "title\r\nbody"

    details = mock_post.call_args_list[3]
    payload = loads(details[1]["data"])
    assert isinstance(payload["to"], dict)
    assert payload["to"]["name"] == "12"
    assert payload["body"] == "title\r\nbody"

    # Verify our URL looks good
    assert obj.url().startswith(
        "bulksms://{}:{}@{}".format(
            user,
            pwd,
            "/".join(["+15551231234", "+15555555555", "@group", "@12"]),
        )
    )

    assert "batch=no" in obj.url()

    # With our batch in place, our calculations are different
    obj = Apprise.instantiate(
        "bulksms://{}:{}@{}?batch=y".format(user, pwd, "/".join(targets))
    )
    # 2 groups and 2 phones are lumped together
    assert len(obj) == 3



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise
from apprise.plugins.voipms import NotifyVoipms

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    (
        "voipms://",
        {
            # No email/password specified
            "instance": TypeError,
        },
    ),
    (
        "voipms://@:",
        {
            # Invalid url
            "instance": TypeError,
        },
    ),
    (
        "voipms://{}/{}".format("user@example.com", "1" * 11),
        {
            # No password specified
            "instance": TypeError,
        },
    ),
    (
        "voipms://:{}".format("password"),
        {
            # No email specified
            "instance": TypeError,
        },
    ),
    (
        "voipms://{}:{}/{}".format("user@", "pass", "1" * 11),
        {
            # Check valid email
            "instance": TypeError,
        },
    ),
    (
        "voipms://{password}:{email}".format(
            email="user@example.com", password="password"
        ),
        {
            # No from_phone specified
            "instance": TypeError,
        },
    ),
    # Invalid phone number test
    (
        "voipms://{password}:{email}/1613".format(
            email="user@example.com", password="password"
        ),
        {
            # Invalid phone number
            "instance": TypeError,
        },
    ),
    # Invalid country code phone number test
    (
        "voipms://{password}:{email}/01133122446688".format(
            email="user@example.com", password="password"
        ),
        {
            # Non North American phone number
            "instance": TypeError,
        },
    ),
    (
        "voipms://{password}:{email}/{from_phone}/{targets}/".format(
            email="user@example.com",
            password="password",
            from_phone="16134448888",
            targets="/".join(["26134442222"]),
        ),
        {
            # Invalid target phone number
            "instance": NotifyVoipms,
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "voipms://{password}:{email}/{from_phone}".format(
            email="user@example.com",
            password="password",
            from_phone="16138884444",
        ),
        {
            "instance": NotifyVoipms,
            # No targets specified
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "voipms://{password}:{email}/?from={from_phone}".format(
            email="user@example.com",
            password="password",
            from_phone="16138884444",
        ),
        {
            "instance": NotifyVoipms,
            # No targets specified
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "voipms://{password}:{email}/{from_phone}/{targets}/".format(
            email="user@example.com",
            password="password",
            from_phone="16138884444",
            targets="/".join(["16134442222"]),
        ),
        {
            # Valid
            "instance": NotifyVoipms,
            "response": True,
            "privacy_url": "voipms://p...d:user@example.com/16...4",
        },
    ),
    (
        "voipms://{password}:{email}/{from_phone}/{targets}/".format(
            email="user@example.com",
            password="password",
            from_phone="16138884444",
            targets="/".join(["16134442222", "16134443333"]),
        ),
        {
            # Valid multiple targets
            "instance": NotifyVoipms,
            "response": True,
            "privacy_url": "voipms://p...d:user@example.com/16...4",
        },
    ),
    (
        "voipms://{password}:{email}/?from={from_phone}&to={targets}".format(
            email="user@example.com",
            password="password",
            from_phone="16138884444",
            targets="16134448888",
        ),
        {
            # Valid
            "instance": NotifyVoipms,
        },
    ),
    (
        "voipms://{password}:{email}/{from_phone}/{targets}/".format(
            email="user@example.com",
            password="password",
            from_phone="16138884444",
            targets="16134442222",
        ),
        {
            "instance": NotifyVoipms,
            # Throws a series of errors
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_voipms():
    """NotifyVoipms() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.get")
def test_plugin_voipms_edge_cases(mock_get):
    """NotifyVoipms() Edge Cases."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_get.return_value = response

    # Initialize some generic (but valid) tokens
    email = "user@example.com"
    password = "password"
    source = "+1 (555) 123-3456"
    targets = "+1 (555) 123-9876"

    # No email specified
    with pytest.raises(TypeError):
        NotifyVoipms(email=None, source=source)

    # a error response is returned
    response.status_code = 400
    response.content = dumps(
        {
            "code": 21211,
            "message": "Unable to process your request.",
        }
    )
    mock_get.return_value = response

    # Initialize our object
    obj = Apprise.instantiate(
        f"voipms://{password}:{email}/{source}/{targets}"
    )

    assert isinstance(obj, NotifyVoipms)

    # We will fail with the above error code
    assert obj.notify("title", "body", "info") is False


@mock.patch("requests.get")
def test_plugin_voipms_non_success_status(mock_get):
    """NotifyVoipms() Non Success Status."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_get.return_value = response

    # A 200 response is returned but non-success message
    response.status_code = 200
    response.content = dumps(
        {
            "status": "invalid_credentials",
            "message": "Username or Password is incorrect",
        }
    )

    obj = Apprise.instantiate(
        "voipms://{password}:{email}/{source}/{targets}".format(
            email="user@example.com",
            password="badpassword",
            source="16134448888",
            targets="16134442222",
        )
    )

    assert isinstance(obj, NotifyVoipms)

    # We will fail with the above error code
    assert obj.notify("title", "body", "info") is False

    response.content = "{"
    assert obj.send("title", "body") is False



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import os
import re
from unittest import mock

from helpers import AppriseURLTester
import requests

from apprise import Apprise, AppriseAttachment, NotifyType
from apprise.plugins.custom_xml import NotifyXML

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")


# Our Testing URLs
apprise_url_tests = (
    (
        "xml://:@/",
        {
            "instance": None,
        },
    ),
    (
        "xml://",
        {
            "instance": None,
        },
    ),
    (
        "xmls://",
        {
            "instance": None,
        },
    ),
    (
        "xml://localhost",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost?method=invalid",
        {
            "instance": TypeError,
        },
    ),
    (
        "xml://user:pass@localhost",
        {
            "instance": NotifyXML,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "xml://user:****@localhost",
        },
    ),
    # Test method variations
    (
        "xml://user@localhost?method=put",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost?method=get",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost?method=post",
        {
            "instance": NotifyXML,
            "force_debug": True,
        },
    ),
    (
        "xml://user@localhost?method=head",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost?method=delete",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost?method=patch",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost?method=update",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user@localhost?method=options",
        {
            "instance": NotifyXML,
        },
    ),
    # Continue testing other cases
    (
        "xml://localhost:8080",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user:pass@localhost:8080",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xmls://localhost",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xmls://user:pass@localhost",
        {
            "instance": NotifyXML,
        },
    ),
    # Continue testing other cases
    (
        "xml://localhost:8080",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user:pass@localhost:8080",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://localhost",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xmls://user:pass@localhost",
        {
            "instance": NotifyXML,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "xmls://user:****@localhost",
        },
    ),
    (
        "xml://user@localhost:8080/path/",
        {
            "instance": NotifyXML,
            "privacy_url": "xml://user@localhost:8080/path",
        },
    ),
    (
        "xmls://localhost:8080/path/",
        {
            "instance": NotifyXML,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "xmls://localhost:8080/path/",
        },
    ),
    (
        "xmls://user:pass@localhost:8080",
        {
            "instance": NotifyXML,
        },
    ),
    # Test our GET params
    (
        "xml://localhost:8080/path?-ParamA=Value",
        {
            "instance": NotifyXML,
        },
    ),
    # Test our Headers
    (
        "xml://localhost:8080/path?+HeaderKey=HeaderValue",
        {
            "instance": NotifyXML,
        },
    ),
    (
        "xml://user:pass@localhost:8081",
        {
            "instance": NotifyXML,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "xml://user:pass@localhost:8082",
        {
            "instance": NotifyXML,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "xml://user:pass@localhost:8083",
        {
            "instance": NotifyXML,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_custom_xml_urls():
    """NotifyXML() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.request")
def test_notify_xml_plugin_attachments(mock_request):
    """NotifyXML() Attachments."""

    okay_response = requests.Request()
    okay_response.status_code = requests.codes.ok
    okay_response.content = ""

    # Assign our mock object our return value
    mock_request.return_value = okay_response

    obj = Apprise.instantiate("xml://localhost.localdomain/")
    assert isinstance(obj, NotifyXML)

    # Test Valid Attachment
    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # Test invalid attachment
    path = os.path.join(TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg")
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=path,
        )
        is False
    )

    # Test Valid Attachment (load 3)
    path = (
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
    )
    attach = AppriseAttachment(path)

    # Return our good configuration
    mock_request.side_effect = None
    mock_request.return_value = okay_response
    with mock.patch("builtins.open", side_effect=OSError()):
        # We can't send the message we can't open the attachment for reading
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )

    # test the handling of our batch modes
    obj = Apprise.instantiate("xml://no-reply@example.com/")
    assert isinstance(obj, NotifyXML)

    # Now send an attachment normally without issues
    mock_request.reset_mock()
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )
    assert mock_request.call_count == 1


@mock.patch("requests.request")
def test_plugin_custom_xml_edge_cases(mock_request):
    """NotifyXML() Edge Cases."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_request.return_value = response

    results = NotifyXML.parse_url(
        "xml://localhost:8080/command?:Message=Body&method=GET"
        "&:Key=value&:,=invalid&:MessageType="
    )

    assert isinstance(results, dict)
    assert results["user"] is None
    assert results["password"] is None
    assert results["port"] == 8080
    assert results["host"] == "localhost"
    assert results["fullpath"] == "/command"
    assert results["path"] == "/"
    assert results["query"] == "command"
    assert results["schema"] == "xml"
    assert results["url"] == "xml://localhost:8080/command"
    assert isinstance(results["qsd:"], dict)
    assert results["qsd:"]["Message"] == "Body"
    assert results["qsd:"]["Key"] == "value"
    assert results["qsd:"][","] == "invalid"

    instance = NotifyXML(**results)
    assert isinstance(instance, NotifyXML)

    # XSD URL is disabled due to custom formatting
    assert instance.xsd_url is None

    response = instance.send(title="title", body="body")
    assert response is True
    assert mock_request.call_count == 1

    details = mock_request.call_args_list[0]
    assert "NotifyType." not in details[1]["data"]
    assert details[0][0] == "GET"
    assert details[0][1] == "http://localhost:8080/command"
    assert instance.url(privacy=False).startswith(
        "xml://localhost:8080/command?"
    )

    # Generate a new URL based on our last and verify key values are the same
    new_results = NotifyXML.parse_url(instance.url(safe=False))
    for k in (
        "user",
        "password",
        "port",
        "host",
        "fullpath",
        "path",
        "query",
        "schema",
        "url",
        "method",
    ):
        assert new_results[k] == results[k]

    # Test our data set for our key/value pair
    assert re.search(r"<Version>[1-9]+\.[0-9]+</Version>", details[1]["data"])
    assert "<Subject>title</Subject>" in details[1]["data"]

    assert "<Message>test</Message>" not in details[1]["data"]
    assert "<Message>" not in details[1]["data"]
    # MessageType was removed from the payload
    assert "<MessageType>" not in details[1]["data"]
    # However we can find our mapped Message to the new value Body
    assert "<Body>body</Body>" in details[1]["data"]
    # Custom entry
    assert "<Key>value</Key>" in details[1]["data"]

    mock_request.reset_mock()

    results = NotifyXML.parse_url(
        "xml://localhost:8081/command?method=POST&:New=Value"
    )

    assert isinstance(results, dict)
    assert results["user"] is None
    assert results["password"] is None
    assert results["port"] == 8081
    assert results["host"] == "localhost"
    assert results["fullpath"] == "/command"
    assert results["path"] == "/"
    assert results["query"] == "command"
    assert results["schema"] == "xml"
    assert results["url"] == "xml://localhost:8081/command"
    assert isinstance(results["qsd:"], dict)
    assert results["qsd:"]["New"] == "Value"

    instance = NotifyXML(**results)
    assert isinstance(instance, NotifyXML)

    # XSD URL is disabled due to custom formatting
    assert instance.xsd_url is None

    response = instance.send(title="title", body="body")
    assert response is True
    assert mock_request.call_count == 1

    details = mock_request.call_args_list[0]
    assert "NotifyType." not in details[1]["data"]
    assert details[0][0] == "POST"
    assert details[0][1] == "http://localhost:8081/command"
    assert instance.url(privacy=False).startswith(
        "xml://localhost:8081/command?"
    )

    # Generate a new URL based on our last and verify key values are the same
    new_results = NotifyXML.parse_url(instance.url(safe=False))
    for k in (
        "user",
        "password",
        "port",
        "host",
        "fullpath",
        "path",
        "query",
        "schema",
        "url",
        "method",
    ):
        assert new_results[k] == results[k]

    # Test our data set for our key/value pair
    assert re.search(r"<Version>[1-9]+\.[0-9]+</Version>", details[1]["data"])
    assert re.search(
        r"<MessageType>{}</MessageType>".format(NotifyType.INFO.value),
        details[1]["data"],
    )
    assert r"<Subject>title</Subject>" in details[1]["data"]
    # No over-ride
    assert r"<Message>body</Message>" in details[1]["data"]

    mock_request.reset_mock()

    results = NotifyXML.parse_url(
        "xmls://localhost?method=POST&:Message=Body&:Subject=Title&:Version"
    )

    assert isinstance(results, dict)
    assert results["user"] is None
    assert results["password"] is None
    assert results["port"] is None
    assert results["host"] == "localhost"
    assert results["fullpath"] is None
    assert results["path"] is None
    assert results["query"] is None
    assert results["schema"] == "xmls"
    assert results["url"] == "xmls://localhost"
    assert isinstance(results["qsd:"], dict)
    assert results["qsd:"]["Version"] == ""
    assert results["qsd:"]["Message"] == "Body"
    assert results["qsd:"]["Subject"] == "Title"

    instance = NotifyXML(**results)
    assert isinstance(instance, NotifyXML)

    # XSD URL is disabled due to custom formatting
    assert instance.xsd_url is None

    response = instance.send(title="title", body="body")
    assert response is True
    assert mock_request.call_count == 1

    details = mock_request.call_args_list[0]
    assert "NotifyType." not in details[1]["data"]
    assert details[0][0] == "POST"
    assert details[0][1] == "https://localhost"
    assert instance.url(privacy=False).startswith("xmls://localhost")

    # Generate a new URL based on our last and verify key values are the same
    new_results = NotifyXML.parse_url(instance.url(safe=False))

    # Test that the Version has been dropped
    assert (
        re.search(r"<Version>[1-9]+\.[0-9]+</Version>", details[1]["data"])
        is None
    )

    # Test our data set for our key/value pair
    assert re.search(
        r"<MessageType>{}</MessageType>".format(NotifyType.INFO.value),
        details[1]["data"],
    )

    # Subject is swapped for Title
    assert r"<Subject>title</Subject>" not in details[1]["data"]
    assert r"<Title>title</Title>" in details[1]["data"]

    # Message is swapped for Body
    assert r"<Message>body</Message>" not in details[1]["data"]
    assert r"<Body>body</Body>" in details[1]["data"]

    mock_request.reset_mock()

    # Verify UPDATE method is forwarded correctly
    results = NotifyXML.parse_url("xml://localhost?method=UPDATE")
    instance = NotifyXML(**results)
    assert instance.send(title="title", body="body") is True
    assert mock_request.call_count == 1
    details = mock_request.call_args_list[0]
    assert details[0][0] == "UPDATE"
    assert details[0][1] == "http://localhost"

    mock_request.reset_mock()

    # Verify OPTIONS method is forwarded correctly
    results = NotifyXML.parse_url("xml://localhost?method=OPTIONS")
    instance = NotifyXML(**results)
    assert instance.send(title="title", body="body") is True
    assert mock_request.call_count == 1
    details = mock_request.call_args_list[0]
    assert details[0][0] == "OPTIONS"
    assert details[0][1] == "http://localhost"



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json

# Disable logging for a cleaner testing output
import logging
import os
import sys
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import asset, exception, url
from apprise.common import PersistentStoreMode
from apprise.plugins.vapid import NotifyVapid
from apprise.plugins.vapid.subscription import (
    WebPushSubscription,
    WebPushSubscriptionManager,
)
from apprise.utils.pem import ApprisePEMController

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# a test UUID we can use
SUBSCRIBER = "user@example.com"

PLUGIN_ID = "vapid"

# Our Testing URLs
apprise_url_tests = (
    (
        "vapid://",
        {
            "instance": TypeError,
        },
    ),
    (
        "vapid://:@/",
        {
            "instance": TypeError,
        },
    ),
    (
        "vapid://invalid-subscriber",
        {
            # An invalid Subscriber
            "instance": TypeError,
        },
    ),
    (
        "vapid://user@example.com",
        {
            # bare bone requirements met, but we don't have our subscription
            # file or our private key (pem)
            "instance": NotifyVapid,
            # We'll fail to respond because we would not have found any
            # configuration to load
            "notify_response": False,
        },
    ),
    (
        "vapid://user@example.com?keyfile=invalid&subfile=invalid",
        {
            # Test passing keyfile and subfile on our path (even if invalid)
            "instance": NotifyVapid,
            # We'll fail to respond because we would not have found any
            # configuration to load
            "notify_response": False,
        },
    ),
    (
        "vapid://user@example.com/newuser@example.com",
        {
            # we don't have our subscription file or private key
            "instance": NotifyVapid,
            "notify_response": False,
        },
    ),
    (
        "vapid://user@example.ca/newuser@example.ca",
        {
            "instance": NotifyVapid,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "vapid://user@example.uk/newuser@example.uk",
        {
            "instance": NotifyVapid,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "vapid://user@example.au/newuser@example.au",
        {
            "instance": NotifyVapid,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


@pytest.fixture
def patch_persistent_store_namespace(tmpdir):
    """Force an easy to test environment."""
    with (
        mock.patch.object(url.URLBase, "url_id", return_value=PLUGIN_ID),
        mock.patch.object(
            asset.AppriseAsset, "storage_mode", PersistentStoreMode.AUTO
        ),
        mock.patch.object(asset.AppriseAsset, "storage_path", str(tmpdir)),
    ):
        tmp_dir = tmpdir.mkdir(PLUGIN_ID)
        # Return the directory name
        yield str(tmp_dir)


@pytest.fixture
def subscription_reference():
    return {
        "user@example.com": {
            "endpoint": "https://fcm.googleapis.com/fcm/send/default",
            "keys": {
                "p256dh": (
                    "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                    "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
                ),
                "auth": "k9Xzm43nBGo=",
            },
        },
        "user1": {
            "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
            "keys": {
                "p256dh": (
                    "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                    "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
                ),
                "auth": "k9Xzm43nBGo=",
            },
        },
        "user2": {
            "endpoint": "https://fcm.googleapis.com/fcm/send/def456",
            "keys": {
                "p256dh": (
                    "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                    "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
                ),
                "auth": "k9Xzm43nBGo=",
            },
        },
    }


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_vapid_urls():
    """
    NotifyVapid() Apprise URLs - No Config

    """

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_vapid_urls_with_required_assets(
    patch_persistent_store_namespace, subscription_reference
):
    """NotifyVapid() Apprise URLs With Config."""

    # Determine our store
    pc = ApprisePEMController(path=patch_persistent_store_namespace)
    assert pc.keygen() is True

    # Write our subscriptions file to disk
    subscription_file = os.path.join(
        patch_persistent_store_namespace, NotifyVapid.vapid_subscription_file
    )

    with open(subscription_file, "w") as f:
        f.write(json.dumps(subscription_reference))

    tests = (
        (
            "vapid://user@example.com",
            {
                # user@example.com loaded (also used as subscriber id)
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user@example.com/newuser@example.com",
            {
                # no newuser@example.com key entry
                "instance": NotifyVapid,
                "notify_response": False,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2",
            {
                # We'll succesfully notify 2 users
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user1?to=user2&from=user@example.com",
            {
                # We'll succesfully notify 2 users
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://?to=user2&from=user@example.com",
            {
                # No host provided
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user@example.com?to=user2&from=user@example.com",
            {
                # We'll succesfully notify 2 users
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2&ttl=15",
            {
                # test ttl
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2&ttl=",
            {
                # test ttl
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2&ttl=invalid",
            {
                # test ttl
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2&ttl=-4000",
            {
                # bad ttl
                "instance": TypeError,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2&mode=edge",
            {
                # test mode
                "instance": NotifyVapid,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2&mode=",
            {
                # test mode
                "instance": TypeError,
            },
        ),
        (
            "vapid://user@example.com/user1?to=user2&mode=invalid",
            {
                # test mode more
                "instance": TypeError,
            },
        ),
        (
            "vapid://user@example.com/user1",
            {
                "instance": NotifyVapid,
                # force a failure
                "response": False,
                "requests_response_code": requests.codes.internal_server_error,
            },
        ),
        (
            "vapid://user@example.com/user1",
            {
                "instance": NotifyVapid,
                # throw a bizarre code forcing us to fail to look it up
                "response": False,
                "requests_response_code": 999,
            },
        ),
        (
            "vapid://user@example.com/user1",
            {
                "instance": NotifyVapid,
                # Throws a series of connection and transfer exceptions
                # when this flag is set and tests that we gracefully handle
                # them
                "test_requests_exceptions": True,
            },
        ),
    )

    AppriseURLTester(tests=tests).run_all()


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_vapid_subscriptions(tmpdir):
    """NotifyVapid() Subscriptions."""

    # Temporary directory
    tmpdir0 = tmpdir.mkdir("tmp00")

    with pytest.raises(exception.AppriseInvalidData):
        # Integer not supported
        WebPushSubscription(42)

    with pytest.raises(exception.AppriseInvalidData):
        # Not the correct format
        WebPushSubscription("bad-content")

    with pytest.raises(exception.AppriseInvalidData):
        # Invalid JSON
        WebPushSubscription("{")

    with pytest.raises(exception.AppriseInvalidData):
        # Empty Dictionary
        WebPushSubscription({})

    with pytest.raises(exception.AppriseInvalidData):
        WebPushSubscription(
            {
                "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
                "keys": {
                    "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR=",
                    "auth": 42,
                },
            }
        )

    with pytest.raises(exception.AppriseInvalidData):
        WebPushSubscription(
            {
                "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
                "keys": {
                    "p256dh": 42,
                    "auth": "k9Xzm43nBGo=",
                },
            }
        )

    with pytest.raises(exception.AppriseInvalidData):
        WebPushSubscription(
            {
                "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
            }
        )

    with pytest.raises(exception.AppriseInvalidData):
        WebPushSubscription(
            {
                "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
                "keys": {},
            }
        )

    with pytest.raises(exception.AppriseInvalidData):
        # Invalid p256dh public key provided
        wps = WebPushSubscription(
            {
                "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
                "keys": {
                    "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR=",
                    "auth": "k9Xzm43nBGo=",
                },
            }
        )

    # An empty object
    wps = WebPushSubscription()
    assert bool(wps) is False
    assert isinstance(wps.json(), str)
    assert json.loads(wps.json())
    assert str(wps) == ""
    assert wps.auth is None
    assert wps.endpoint is None
    assert wps.p256dh is None
    assert wps.public_key is None
    # We can't write anything as there is nothing loaded
    assert wps.write(os.path.join(str(tmpdir0), "subscriptions.json")) is False

    # A valid key
    wps = WebPushSubscription(
        {
            "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
            "keys": {
                "p256dh": (
                    "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                    "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
                ),
                "auth": "k9Xzm43nBGo=",
            },
        }
    )

    assert bool(wps) is True
    assert isinstance(wps.json(), str)
    assert json.loads(wps.json())
    assert str(wps) == "abc123"
    assert wps.auth == "k9Xzm43nBGo="
    assert wps.endpoint == "https://fcm.googleapis.com/fcm/send/abc123"
    assert (
        wps.p256dh == "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
        "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
    )
    assert wps.public_key is not None

    # Currently no files here
    assert os.listdir(str(tmpdir0)) == []

    # Bad content
    assert wps.write(object) is False
    assert wps.write(None) is False
    # Can't write to a name already taken by as a directory
    assert wps.write(str(tmpdir0)) is False
    # Can't write to a name already taken by as a directory
    assert wps.write(os.path.join(str(tmpdir0), "subscriptions.json")) is True
    assert os.listdir(str(tmpdir0)) == ["subscriptions.json"]


@pytest.mark.skipif(
    "cryptography" in sys.modules,
    reason="Requires that cryptography NOT be installed",
)
def test_plugin_vapid_subscriptions_without_c():
    """NotifyVapid() Subscriptions (no Cryptography)"""
    with pytest.raises(exception.AppriseInvalidData):
        # A valid key that can't be loaded because crytography is missing
        WebPushSubscription(
            {
                "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
                "keys": {
                    "p256dh": (
                        "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                        "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
                    ),
                    "auth": "k9Xzm43nBGo=",
                },
            }
        )


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_vapid_subscription_manager(tmpdir):
    """NotifyVapid() Subscription Manager."""

    # Temporary directory
    tmpdir0 = tmpdir.mkdir("tmp00")

    with pytest.raises(exception.AppriseInvalidData):
        # An invalid object
        smgr = WebPushSubscriptionManager()
        smgr["abc"] = "invalid"

    with pytest.raises(exception.AppriseInvalidData):
        # An invalid object
        smgr = WebPushSubscriptionManager()
        smgr += "invalid"

    smgr = WebPushSubscriptionManager()

    assert bool(smgr) is False
    assert len(smgr) == 0

    sub = {
        "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
        "keys": {
            "p256dh": (
                "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
            ),
            "auth": "k9Xzm43nBGo=",
        },
    }

    assert smgr.add(sub) is True
    assert bool(smgr) is True
    assert len(smgr) == 1

    # Same sub (overwrites same slot)
    smgr += sub
    assert bool(smgr) is True
    assert len(smgr) == 1

    # This makes a copy
    smgr["abc"] = smgr["abc123"]
    assert bool(smgr) is True
    assert len(smgr) == 2

    assert isinstance(smgr["abc123"], WebPushSubscription)

    # Currently no files here
    assert os.listdir(str(tmpdir0)) == []

    # Write our content
    assert smgr.write(os.path.join(str(tmpdir0), "subscriptions.json")) is True

    assert os.listdir(str(tmpdir0)) == ["subscriptions.json"]

    # Reset our object
    smgr.clear()
    assert bool(smgr) is False
    assert len(smgr) == 0

    # Load our content back
    assert smgr.load(os.path.join(str(tmpdir0), "subscriptions.json")) is True
    assert bool(smgr) is True
    assert len(smgr) == 2

    # Write over our file using the standard Subscription format
    assert (
        smgr["abc123"].write(os.path.join(str(tmpdir0), "subscriptions.json"))
        is True
    )

    # We can still open this type as well
    assert smgr.load(os.path.join(str(tmpdir0), "subscriptions.json")) is True
    assert bool(smgr) is True
    assert len(smgr) == 1

    smgr.clear()
    bad_entry = {
        "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
        "keys": {
            "p256dh": "invalid",
            "auth": "garbage",
        },
    }

    subscriptions = os.path.join(str(tmpdir0), "subscriptions.json")
    with open(subscriptions, "w", encoding="utf-8") as f:
        # A bad JSON file
        f.write("{")
    assert smgr.load(subscriptions) is False

    with open(subscriptions, "w", encoding="utf-8") as f:
        # not expected dictionary
        f.write("null")
    assert smgr.load(subscriptions) is False

    subscriptions = os.path.join(str(tmpdir0), "subscriptions.json")
    with open(subscriptions, "w", encoding="utf-8") as f:
        json.dump(bad_entry, f)
    assert smgr.load(subscriptions) is False

    # Create bad data
    bad_data = {
        "bad1": bad_entry,
        "bad2": bad_entry,
        "bad3": bad_entry,
        "bad4": bad_entry,
    }
    subscriptions = os.path.join(str(tmpdir0), "subscriptions.json")
    with open(subscriptions, "w", encoding="utf-8") as f:
        json.dump(bad_data, f)
    assert smgr.load(subscriptions) is False
    assert smgr.load("invalid-file") is False


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
@mock.patch("requests.post")
def test_plugin_vapid_initializations(mock_post, tmpdir):
    """NotifyVapid() Initializations."""

    # Assign our mock object our return value
    okay_response = requests.Request()
    okay_response.status_code = requests.codes.ok
    okay_response.content = ""
    mock_post.return_value = okay_response

    # Temporary directory
    tmpdir0 = tmpdir.mkdir("tmp00")

    # Write our subfile
    smgr = WebPushSubscriptionManager()
    sub = {
        "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
        "keys": {
            "p256dh": (
                "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
            ),
            "auth": "k9Xzm43nBGo=",
        },
    }
    subfile = os.path.join(str(tmpdir0), "subscriptions.json")
    assert smgr.add(sub) is True
    assert smgr.add(smgr["abc123"]) is True
    assert os.listdir(str(tmpdir0)) == []

    with mock.patch("json.dump", side_effect=OSError):
        # We will fial to write
        assert smgr.write(subfile) is False

    assert smgr.write(subfile) is True
    assert os.listdir(str(tmpdir0)) == ["subscriptions.json"]
    assert isinstance(smgr.json(), str)

    asset_ = asset.AppriseAsset(
        storage_mode=PersistentStoreMode.FLUSH,
        storage_path=str(tmpdir0),
        # Auto-gen our private/public key pair
        pem_autogen=True,
    )

    # Auto-Key Generation
    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        subfile=subfile,
        asset=asset_,
    )
    assert isinstance(obj, NotifyVapid)
    # Our subscription directory + our
    # persistent store where our keys were generated
    assert len(os.listdir(str(tmpdir0))) == 2

    # Second call re-references keys previously generated
    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        subfile=subfile,
        asset=asset_,
    )
    assert isinstance(obj, NotifyVapid)
    assert isinstance(obj.url(), str)
    assert obj.send("test") is True
    # A second message makes no difference; what is loaded into memory is used
    assert obj.send("test") is True

    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        subfile="/a/bad/path",
        asset=asset_,
    )
    assert isinstance(obj, NotifyVapid)
    assert isinstance(obj.url(), str)
    assert obj.send("test") is False
    # A second message makes no difference; what is loaded into memory is used
    assert obj.send("test") is False

    # Detect our keyfile
    cache_dir = next(
        x
        for x in os.listdir(str(tmpdir0))
        if not x.endswith("subscriptions.json")
    )

    # Test fixed assignment to our keyfile
    keyfile = os.path.join(str(tmpdir0), cache_dir, "private_key.pem")
    assert os.path.exists(keyfile)
    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        keyfile=keyfile,
        subfile=subfile,
        asset=asset_,
    )
    assert isinstance(obj, NotifyVapid)
    assert isinstance(obj.url(), str)
    assert obj.send("test") is True
    # A second message makes no difference; what is loaded into memory is used
    assert obj.send("test") is True

    # Invalid Keyfile
    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        keyfile=subfile,
        subfile=subfile,
        asset=asset_,
    )
    assert isinstance(obj, NotifyVapid)
    assert isinstance(obj.url(), str)
    assert obj.send("test") is False
    # A second message makes no difference; what is loaded into memory is used
    assert obj.send("test") is False

    # AutoGen Temporary directory
    tmpdir1 = tmpdir.mkdir("tmp01")
    asset2 = asset.AppriseAsset(
        storage_mode=PersistentStoreMode.FLUSH,
        storage_path=str(tmpdir1),
        # Auto-gen our private/public key pair
        pem_autogen=True,
    )

    assert os.listdir(str(tmpdir1)) == []
    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        keyfile=keyfile,
        asset=asset2,
    )
    assert isinstance(obj, NotifyVapid)
    assert isinstance(obj.url(), str)
    # We have a temporary subscription file we can use
    assert os.listdir(str(tmpdir1)) == ["00088ad3"]
    # We will have a dud configuration file, but at least it's something
    # to help the user with
    assert obj.send("test") is False
    # Second instance fails as well
    assert obj.send("test") is False

    # AutoGen Temporary directory
    tmpdir2 = tmpdir.mkdir("tmp02")
    asset3 = asset.AppriseAsset(
        storage_mode=PersistentStoreMode.FLUSH,
        storage_path=str(tmpdir2),
        # Auto-gen our private/public key pair
        pem_autogen=True,
    )

    # Test invalid keyfile
    assert os.path.exists(keyfile)
    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        keyfile="invalid-file",
        subfile=subfile,
        asset=asset3,
    )
    assert isinstance(obj, NotifyVapid)
    assert isinstance(obj.url(), str)
    assert obj.send("test") is False
    # A second message makes no difference; what is loaded into memory is used
    assert obj.send("test") is False


@pytest.mark.skipif(
    "cryptography" in sys.modules,
    reason="Requires that cryptography NOT be installed",
)
def test_plugin_vapid_initializations_without_c(tmpdir):
    """NotifyVapid() Initializations without cryptography."""
    # Temporary directory
    tmpdir0 = tmpdir.mkdir("tmp00")

    # Write our subfile
    smgr = WebPushSubscriptionManager()
    sub = {
        "endpoint": "https://fcm.googleapis.com/fcm/send/abc123",
        "keys": {
            "p256dh": (
                "BI2RNIK2PkeCVoEfgVQNjievBi4gWvZxMiuCpOx6K6qCO"
                "5caru5QCPuc-nEaLplbbFkHxTrR9YzE8ZkTjie5Fq0"
            ),
            "auth": "k9Xzm43nBGo=",
        },
    }
    subfile = os.path.join(str(tmpdir0), "subscriptions.json")
    assert smgr.add(sub) is False
    asset_ = asset.AppriseAsset(
        storage_mode=PersistentStoreMode.FLUSH,
        storage_path=str(tmpdir0),
        # Auto-gen our private/public key pair
        pem_autogen=True,
    )

    # Auto-Key Generation
    obj = NotifyVapid(
        "user@example.ca",
        targets=[
            "abc123",
        ],
        subfile=subfile,
        asset=asset_,
    )
    assert isinstance(obj, NotifyVapid)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import sys
from unittest import mock

import pytest

from apprise import AppriseAsset, PersistentStoreMode, utils

# Disable logging for a cleaner testing output
logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_utils_pem_general(tmpdir):
    """Utils:PEM."""

    # string to manipulate/work with
    unencrypted_str = "message"

    tmpdir0 = tmpdir.mkdir("tmp00")

    # Currently no files here
    assert os.listdir(str(tmpdir0)) == []

    asset = AppriseAsset(
        storage_mode=PersistentStoreMode.MEMORY,
        storage_path=str(tmpdir0),
        pem_autogen=False,
    )

    # Create a PEM Controller
    pem_c = utils.pem.ApprisePEMController(path=None, asset=asset)

    # Nothing to lookup
    assert pem_c.public_keyfile() is None
    assert pem_c.public_key() is None
    assert pem_c.x962_str == ""
    assert pem_c.decrypt(b"data") is None
    assert pem_c.encrypt(unencrypted_str) is None
    # Keys can not be generated in memory mode
    assert pem_c.keygen() is False
    assert pem_c.sign(b"data") is None

    asset = AppriseAsset(
        storage_mode=PersistentStoreMode.FLUSH,
        storage_path=str(tmpdir0),
        pem_autogen=False,
    )

    # No new files
    assert os.listdir(str(tmpdir0)) == []

    # Our asset is now write mode, so we will be able to generate a key
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    # Nothing to lookup
    assert pem_c.public_keyfile() is None
    assert pem_c.public_key() is None
    assert pem_c.x962_str == ""
    assert pem_c.encrypt(unencrypted_str) is None

    # Generate our keys
    assert bool(pem_c) is False
    assert pem_c.keygen() is True
    assert bool(pem_c) is True

    # We have 2 new key files generated
    pub_keyfile = os.path.join(str(tmpdir0), "public_key.pem")
    prv_keyfile = os.path.join(str(tmpdir0), "private_key.pem")
    assert os.path.isfile(pub_keyfile)
    assert os.path.isfile(prv_keyfile)
    assert pem_c.public_keyfile() is not None
    assert pem_c.decrypt("garbage") is None
    assert pem_c.public_key() is not None

    # Keys used later on as ref
    pubkey_ref = pem_c.public_key()
    prvkey_ref = pem_c.private_key()

    assert isinstance(pem_c.x962_str, str)
    assert len(pem_c.x962_str) > 20
    content = pem_c.encrypt(unencrypted_str)
    assert pem_c.decrypt(
        pem_c.encrypt(unencrypted_str.encode("utf-8"))
    ) == pem_c.decrypt(pem_c.encrypt(unencrypted_str))
    assert pem_c.decrypt(
        pem_c.encrypt(unencrypted_str, public_key=pem_c.public_key())
    ) == pem_c.decrypt(pem_c.encrypt(unencrypted_str))
    assert pem_c.decrypt(content) == unencrypted_str
    assert isinstance(content, str)
    assert pem_c.decrypt(content) == unencrypted_str
    # support str as well
    assert pem_c.decrypt(content) == unencrypted_str
    assert pem_c.decrypt(content.encode("utf-8")) == unencrypted_str
    # Sign test
    assert isinstance(pem_c.sign(content.encode("utf-8")), bytes)

    # Web Push handling
    webpush_content = pem_c.encrypt_webpush(
        unencrypted_str, public_key=pem_c.public_key(), auth_secret=b"secret"
    )
    assert isinstance(webpush_content, bytes)

    webpush_content = pem_c.encrypt_webpush(
        unencrypted_str.encode("utf-8"),
        public_key=pem_c.public_key(),
        auth_secret=b"secret",
    )
    assert isinstance(webpush_content, bytes)

    # Non Bytes (garbage basically)
    with pytest.raises(TypeError):
        assert pem_c.decrypt(None) is None

    with pytest.raises(TypeError):
        assert pem_c.decrypt(5) is None

    with pytest.raises(TypeError):
        assert pem_c.decrypt(False) is None

    with pytest.raises(TypeError):
        assert pem_c.decrypt(object) is None

    # Test our initialization
    pem_c = utils.pem.ApprisePEMController(
        path=None, prv_keyfile="invalid", asset=asset
    )
    assert pem_c.private_keyfile() is False
    assert pem_c.public_keyfile() is None
    assert pem_c.prv_keyfile is False
    assert pem_c.pub_keyfile is None
    assert pem_c.private_key() is None
    assert pem_c.public_key() is None
    assert pem_c.decrypt(content) is None

    pem_c = utils.pem.ApprisePEMController(
        path=None, pub_keyfile="invalid", asset=asset
    )
    assert pem_c.private_keyfile() is None
    assert pem_c.public_keyfile() is False
    assert pem_c.prv_keyfile is None
    assert pem_c.pub_keyfile is False
    assert pem_c.private_key() is None
    assert pem_c.public_key() is None
    assert pem_c.decrypt(content) is None

    pem_c = utils.pem.ApprisePEMController(
        path=None, prv_keyfile=prv_keyfile, asset=asset
    )
    assert pem_c.private_keyfile() == prv_keyfile
    assert pem_c.public_keyfile() is None
    assert pem_c.private_key() is not None
    assert pem_c.prv_keyfile == prv_keyfile
    assert pem_c.pub_keyfile is None
    assert pem_c.public_key() is not None
    assert pem_c.decrypt(content) == unencrypted_str

    pem_c = utils.pem.ApprisePEMController(
        path=None, pub_keyfile=pub_keyfile, asset=asset
    )
    assert pem_c.private_keyfile() is None
    assert pem_c.public_keyfile() == pub_keyfile
    assert pem_c.prv_keyfile is None
    assert pem_c.pub_keyfile == pub_keyfile
    assert pem_c.private_key() is None
    assert pem_c.public_key() is not None
    assert pem_c.decrypt(content) is None

    # Test our path references
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    assert pem_c.load_private_key(path=None) is True
    assert pem_c.private_keyfile() == prv_keyfile
    assert pem_c.prv_keyfile is None
    assert pem_c.pub_keyfile is None
    assert pem_c.decrypt(content) == unencrypted_str

    # Generate a new key referencing another location
    pem_c = utils.pem.ApprisePEMController(
        name="keygen-tests", path=str(tmpdir0), asset=asset
    )

    # generate ourselves some keys
    assert pem_c.keygen() is True
    keygen_prv_file = pem_c.prv_keyfile
    keygen_pub_file = pem_c.pub_keyfile

    # Remove 1 (but not both)
    os.unlink(keygen_pub_file)

    pem_c = utils.pem.ApprisePEMController(
        name="keygen-tests", path=str(tmpdir0), asset=asset
    )
    # Private key was found, so this does not work
    assert pem_c.keygen() is False
    os.unlink(keygen_prv_file)

    pem_c = utils.pem.ApprisePEMController(
        name="keygen-tests", path=str(tmpdir0), asset=asset
    )
    # It works now
    assert pem_c.keygen() is True

    # Tests public_key generation failure only
    with mock.patch("builtins.open", side_effect=OSError()):
        assert pem_c.keygen(force=True) is False
        with mock.patch("os.unlink", side_effect=OSError()):
            assert pem_c.keygen(force=True) is False
        with mock.patch("os.unlink", return_value=True):
            assert pem_c.keygen(force=True) is False

    # Tests private key generation
    side_effect = [mock.mock_open(read_data="file contents").return_value] + [
        OSError() for _ in range(10)
    ]
    with mock.patch("builtins.open", side_effect=side_effect):
        assert pem_c.keygen(force=True) is False
    with (
        mock.patch("builtins.open", side_effect=side_effect),
        mock.patch("os.unlink", side_effect=OSError()),
    ):
        assert pem_c.keygen(force=True) is False
    with (
        mock.patch("builtins.open", side_effect=side_effect),
        mock.patch("os.unlink", return_value=True),
    ):
        assert pem_c.keygen(force=True) is False

    # Generate a new key referencing another location
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    # We can't re-generate keys if ones already exist
    assert pem_c.keygen() is False
    # the keygen is the big difference here
    assert pem_c.keygen(name="test") is True
    # under the hood, a key is not regenerated (as one already exists)
    assert pem_c.keygen(name="test") is False
    # Generate it a second time by force
    assert pem_c.keygen(name="test", force=True) is True

    assert pem_c.private_keyfile() == os.path.join(
        str(tmpdir0), "test-private_key.pem"
    )
    assert pem_c.public_keyfile() == os.path.join(
        str(tmpdir0), "test-public_key.pem"
    )
    assert pem_c.private_key() is not None
    assert pem_c.public_key() is not None
    assert pem_c.prv_keyfile == os.path.join(
        str(tmpdir0), "test-private_key.pem"
    )
    assert pem_c.pub_keyfile == os.path.join(
        str(tmpdir0), "test-public_key.pem"
    )
    # 'content' was generated using a different key and can not be
    # decrypted
    assert pem_c.decrypt(content) is None

    # Test Decryption files
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    # Calling decrypt triggers underlining code to auto-load
    assert pem_c.decrypt(content) == unencrypted_str
    # Using a private key by path
    assert (
        pem_c.decrypt(content, private_key=pem_c.private_key())
        == unencrypted_str
    )

    # Test different edge cases of load_private_key()
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    assert pem_c.load_private_key() is True
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    assert pem_c.load_private_key(path=prv_keyfile) is True
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    with mock.patch("builtins.open", side_effect=TypeError()):
        assert pem_c.load_private_key(path=prv_keyfile) is False
    with mock.patch("builtins.open", side_effect=OSError()):
        assert pem_c.load_private_key(path=prv_keyfile) is False
    with mock.patch("builtins.open", side_effect=FileNotFoundError()):
        assert pem_c.load_private_key(path=prv_keyfile) is False

    # Test different edge cases of load_public_key()
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    assert pem_c.load_public_key() is True
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    assert pem_c.load_public_key(path=pub_keyfile) is True
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    with mock.patch("builtins.open", side_effect=TypeError()):
        assert pem_c.load_public_key(path=pub_keyfile) is False
    with mock.patch("builtins.open", side_effect=OSError()):
        assert pem_c.load_public_key(path=pub_keyfile) is False
    with mock.patch("builtins.open", side_effect=FileNotFoundError()):
        assert pem_c.load_public_key(path=pub_keyfile) is False

    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    assert pem_c.public_keyfile("test1", "test2") == pub_keyfile
    assert pem_c.private_keyfile("test1", "test2") == prv_keyfile

    pem_c = utils.pem.ApprisePEMController(
        path=str(tmpdir0), name="pub1", asset=asset
    )
    assert pem_c.public_key(autogen=True)

    pem_c = utils.pem.ApprisePEMController(
        path=str(tmpdir0), name="pub2", asset=asset
    )
    assert pem_c.private_key(autogen=True)

    #
    # Auto key generation turned on
    #
    asset = AppriseAsset(
        storage_mode=PersistentStoreMode.MEMORY,
        storage_path=str(tmpdir0),
        pem_autogen=True,
    )
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    assert pem_c.load_public_key(path=pub_keyfile) is True
    pem_c = utils.pem.ApprisePEMController(path=None, asset=asset)
    assert pem_c.load_public_key(path=pub_keyfile) is True

    tmpdir1 = tmpdir.mkdir("tmp01")

    # Currently no files here
    assert os.listdir(str(tmpdir1)) == []

    asset = AppriseAsset(
        storage_mode=PersistentStoreMode.MEMORY,
        storage_path=str(tmpdir1),
        pem_autogen=False,
    )

    # Auto-Gen is turned off, so weare not successful here
    pem_c = utils.pem.ApprisePEMController(path=None, asset=asset)
    assert pem_c.public_key() is None
    assert pem_c.private_key() is None
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir1), asset=asset)
    assert pem_c.public_key() is None
    assert pem_c.private_key() is None

    asset = AppriseAsset(
        storage_mode=PersistentStoreMode.FLUSH,
        storage_path=str(tmpdir1),
        pem_autogen=True,
    )
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir1), asset=asset)
    # Generate ourselves a private key
    assert pem_c.public_key() is not None
    assert pem_c.private_key() is not None
    pub_keyfile = os.path.join(str(tmpdir1), "public_key.pem")
    prv_keyfile = os.path.join(str(tmpdir1), "private_key.pem")
    assert os.path.isfile(pub_keyfile)
    assert os.path.isfile(prv_keyfile)

    with open(pub_keyfile, "w") as f:
        f.write("garbage")

    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir1), asset=asset)
    # we can still load our data as the public key is generated
    # from the private
    assert pem_c.public_key() is not None
    assert pem_c.private_key() is not None

    tmpdir2 = tmpdir.mkdir("tmp02")
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir2), asset=asset)
    pub_keyfile = os.path.join(str(tmpdir2), "public_key.pem")
    prv_keyfile = os.path.join(str(tmpdir2), "private_key.pem")
    assert not os.path.isfile(pub_keyfile)
    assert not os.path.isfile(prv_keyfile)

    #
    # Public Key Edge Case Tests
    #
    with (
        mock.patch.object(
            pem_c, "public_keyfile", side_effect=[None, pub_keyfile]
        ) as mock_keyfile,
        mock.patch.object(
            pem_c,
            "keygen",
            side_effect=lambda *_, **__: (
                setattr(pem_c, "_ApprisePEMController__public_key", pubkey_ref)
                or True
            ),
        ) as mock_keygen,
        mock.patch.object(pem_c, "load_public_key", return_value=True),
    ):
        result = pem_c.public_key()
        assert result is pubkey_ref
        assert mock_keyfile.call_count == 2
        mock_keygen.assert_called_once()

    # - First call: None -> triggers keygen
    # - Second call (recursive): None -> causes fallback
    public_keyfile_side_effect = [None, None]

    with (
        mock.patch.object(
            pem_c, "public_keyfile", side_effect=public_keyfile_side_effect
        ) as mock_keyfile,
        mock.patch.object(pem_c, "keygen", return_value=True) as mock_keygen,
        mock.patch.object(
            pem_c, "load_public_key", return_value=False
        ) as mock_load,
    ):
        # Ensure no key is preset initially
        pem_c._ApprisePEMController__public_key = None

        result = pem_c.public_key()
        assert result is None
        # Once in outer call, once in recursive
        assert mock_keyfile.call_count == 2
        mock_keygen.assert_called_once()
        mock_load.assert_not_called()

    #
    # Private Key Edge Case Tests
    #
    with (
        mock.patch.object(
            pem_c, "private_keyfile", side_effect=[None, prv_keyfile]
        ) as mock_keyfile,
        mock.patch.object(
            pem_c,
            "keygen",
            side_effect=lambda *_, **__: (
                setattr(
                    pem_c, "_ApprisePEMController__private_key", prvkey_ref
                )
                or True
            ),
        ) as mock_keygen,
        mock.patch.object(pem_c, "load_private_key", return_value=True),
    ):
        result = pem_c.private_key()
        assert result is prvkey_ref
        assert mock_keyfile.call_count == 2
        mock_keygen.assert_called_once()

    # - First call: None -> triggers keygen
    # - Second call (recursive): None -> causes fallback
    private_keyfile_side_effect = [None, None]

    with (
        mock.patch.object(
            pem_c, "private_keyfile", side_effect=private_keyfile_side_effect
        ) as mock_keyfile,
        mock.patch.object(pem_c, "keygen", return_value=True) as mock_keygen,
        mock.patch.object(
            pem_c, "load_private_key", return_value=False
        ) as mock_load,
    ):
        # Ensure no key is preset initially
        pem_c._ApprisePEMController__private_key = None

        result = pem_c.private_key()
        assert result is None
        # Once in outer call, once in recursive
        assert mock_keyfile.call_count == 2
        mock_keygen.assert_called_once()
        mock_load.assert_not_called()


@pytest.mark.skipif(
    "cryptography" in sys.modules,
    reason="Requires that cryptography NOT be installed",
)
def test_utils_pem_general_without_c(tmpdir):
    """Utils:PEM Without cryptography."""

    tmpdir0 = tmpdir.mkdir("tmp00")

    # Currently no files here
    assert os.listdir(str(tmpdir0)) == []

    asset = AppriseAsset(
        storage_mode=PersistentStoreMode.MEMORY,
        storage_path=str(tmpdir0),
        pem_autogen=False,
    )

    # Create a PEM Controller
    pem_c = utils.pem.ApprisePEMController(path=None, asset=asset)

    # cryptography library missing poses issues with library useage
    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.public_keyfile()

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.public_key()

    with pytest.raises(utils.pem.ApprisePEMException):
        _ = pem_c.x962_str

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.encrypt("message")

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.keygen()

    asset = AppriseAsset(
        storage_mode=PersistentStoreMode.FLUSH,
        storage_path=str(tmpdir0),
        pem_autogen=False,
    )

    # No new files
    assert os.listdir(str(tmpdir0)) == []

    # Our asset is now write mode, so we will be able to generate a key
    pem_c = utils.pem.ApprisePEMController(path=str(tmpdir0), asset=asset)
    # Nothing to lookup
    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.private_keyfile()

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.private_key()

    with pytest.raises(utils.pem.ApprisePEMException):
        _ = pem_c.x962_str

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.encrypt("message")

    # Keys can not be generated in memory mode
    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.keygen()

    # No files loaded
    assert os.listdir(str(tmpdir0)) == []

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.public_keyfile()

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.public_key()

    with pytest.raises(utils.pem.ApprisePEMException):
        _ = pem_c.x962_str

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.encrypt("message")

    with pytest.raises(utils.pem.ApprisePEMException):
        pem_c.decrypt("abcd==")



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime, timezone as _tz, tzinfo
from inspect import cleandoc

# Disable logging for a cleaner testing output
import logging
from typing import Any
from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture
import requests
import yaml

from apprise import Apprise, AppriseAsset, AppriseConfig, ConfigFormat
from apprise.config import ConfigBase
from apprise.plugins.email import NotifyEmail
from apprise.utils.time import zoneinfo

logging.disable(logging.CRITICAL)


@pytest.fixture
def requests_remote_config(mocker: MockerFixture) -> Mock:
    """
    Patch requests.post globally.

    The config loader will still go through its normal HTTP logic, but all
    outbound GETs will receive controlled in-memory responses.
    """

    def fake_post(url: str, *args: Any, **kwargs: Any) -> requests.Response:
        if url == "http://localhost:8000/get/test-001":
            body = cleandoc("""
                json://localhost
                form://localhost
                """)
        elif url == "http://localhost:8000/get/test-002":
            body = cleandoc("""
                xml://localhost
                """)
        else:
            pytest.fail(f"Unexpected URL fetched: {url!r}")

        resp = requests.Response()
        resp.status_code = requests.codes.ok
        resp.url = url
        resp._content = body.encode("utf-8")  # type: ignore[attr-defined]
        resp.encoding = "utf-8"
        return resp

    # Patch the actual requests.post symbol that ConfigHTTP uses internally
    mock_post: Mock = mocker.patch("requests.post", side_effect=fake_post)
    return mock_post


def test_config_base():
    """
    API: ConfigBase() object

    """

    # invalid types throw exceptions
    with pytest.raises(TypeError):
        ConfigBase(**{"format": "invalid"})

    # Config format types are not the same as ConfigBase ones
    with pytest.raises(TypeError):
        ConfigBase(**{"format": "markdown"})

    cb = ConfigBase(**{"format": "yaml"})
    assert isinstance(cb, ConfigBase)

    cb = ConfigBase(**{"format": "text"})
    assert isinstance(cb, ConfigBase)

    # Set encoding
    cb = ConfigBase(encoding="utf-8", format="text")
    assert isinstance(cb, ConfigBase)

    # read is not supported in the base object; only the children
    assert cb.read() is None

    # There are no servers loaded on a freshly created object
    assert len(cb.servers()) == 0

    # Unsupported URLs are not parsed
    assert ConfigBase.parse_url(url="invalid://") is None

    # Valid URL & Valid Format
    results = ConfigBase.parse_url(
        url="file://relative/path?format=yaml&encoding=latin-1"
    )
    assert isinstance(results, dict)
    # These are moved into the root
    assert results.get("format") == "yaml"
    assert results.get("encoding") == "latin-1"

    # But they also exist in the qsd location
    assert isinstance(results.get("qsd"), dict)
    assert results["qsd"].get("encoding") == "latin-1"
    assert results["qsd"].get("format") == "yaml"

    # Valid URL & Invalid Format
    results = ConfigBase.parse_url(
        url="file://relative/path?format=invalid&encoding=latin-1"
    )
    assert isinstance(results, dict)
    # Only encoding is moved into the root
    assert "format" not in results
    assert results.get("encoding") == "latin-1"

    # But they will always exist in the qsd location
    assert isinstance(results.get("qsd"), dict)
    assert results["qsd"].get("encoding") == "latin-1"
    assert results["qsd"].get("format") == "invalid"


def test_config_base_detect_config_format():
    """
    API: ConfigBase.detect_config_format

    """

    # Garbage Handling
    for garbage in (object(), None, 42):
        # A response is always correctly returned
        assert ConfigBase.detect_config_format(garbage) is None

    # Empty files are valid
    assert ConfigBase.detect_config_format("") is ConfigFormat.TEXT

    # Valid Text Configuration
    assert (
        ConfigBase.detect_config_format("""
    # A comment line over top of a URL
    mailto://userb:pass@gmail.com
    """)
        is ConfigFormat.TEXT
    )

    # A text file that has semi-colon as comment characters
    # is valid too
    assert (
        ConfigBase.detect_config_format("""
    ; A comment line over top of a URL
    mailto://userb:pass@gmail.com
    """)
        is ConfigFormat.TEXT
    )

    # Valid YAML Configuration
    assert (
        ConfigBase.detect_config_format("""
    # A comment line over top of a URL
    version: 1
    """)
        is ConfigFormat.YAML
    )

    # Just a whole lot of blank lines...
    assert ConfigBase.detect_config_format("\n\n\n") is ConfigFormat.TEXT

    # Invalid Config
    assert ConfigBase.detect_config_format("3") is None


def test_config_base_config_parse():
    """
    API: ConfigBase.config_parse

    """

    # Garbage Handling
    for garbage in (object(), None, 42):
        # A response is always correctly returned
        result = ConfigBase.config_parse(garbage)
        # response is a tuple...
        assert isinstance(result, tuple)
        # containing 2 items (plugins, config)
        assert len(result) == 2
        # In the case of garbage in, we get garbage out; both lists are empty
        assert result == ([], [])

    # Valid Text Configuration
    result = ConfigBase.config_parse(
        """
    # A comment line over top of a URL
    mailto://userb:pass@gmail.com
    """,
        asset=AppriseAsset(),
    )
    # We expect to parse 1 entry from the above
    assert isinstance(result, tuple)
    assert len(result) == 2
    # The first element is the number of notification services processed
    assert len(result[0]) == 1
    # If we index into the item, we can check to see the tags associate
    # with it
    assert len(result[0][0].tags) == 0

    # The second is the number of configuration include lines parsed
    assert len(result[1]) == 0

    # Valid Configuration
    result = ConfigBase.config_parse(
        """
# if no version is specified then version 1 is presumed
version: 1

#
# Define your notification urls:
#
urls:
  - pbul://o.gn5kj6nfhv736I7jC3cj3QLRiyhgl98b
  - mailto://test:password@gmail.com
  - json://localhost:
      - tag: devops, admin
    """,
        asset=AppriseAsset(),
    )

    # We expect to parse 2 entries from the above
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert isinstance(result[0], list)
    assert len(result[0]) == 3
    assert len(result[0][0].tags) == 0
    assert len(result[0][1].tags) == 0
    assert len(result[0][2].tags) == 2

    # Test case where we pass in a bad format
    result = ConfigBase.config_parse(
        """
    ; A comment line over top of a URL
    mailto://userb:pass@gmail.com
    """,
        config_format="invalid-format",
    )

    # This is not parseable despite the valid text
    assert isinstance(result, tuple)
    assert isinstance(result[0], list)
    assert len(result[0]) == 0

    result, _ = ConfigBase.config_parse(
        """
    ; A comment line over top of a URL
    mailto://userb:pass@gmail.com
    """,
        config_format=ConfigFormat.TEXT,
    )

    # Parseable
    assert isinstance(result, list)
    assert len(result) == 1


def test_config_base_discord_bug_report_01():
    """
    API: ConfigBase.config_parse user feedback

    A Discord report that a tag was not correctly assigned to a URL when
    presented in the following format
       urls:
         - json://myhost:
           - tag: test
             userid: test
    """
    result, config = ConfigBase.config_parse(
        """
    urls:
      - json://myhost:
        - tag: test
          userid: test
    """,
        asset=AppriseAsset(),
    )

    # We expect to parse 4 entries from the above
    assert isinstance(result, list)
    assert isinstance(config, list)
    assert len(result) == 1
    assert len(result[0].tags) == 1
    assert "test" in result[0].tags


def test_config_base_config_parse_text():
    """
    API: ConfigBase.config_parse_text object

    """

    # Garbage Handling
    for garbage in (object(), None, 42):
        # A response is always correctly returned
        result = ConfigBase.config_parse_text(garbage)
        # response is a tuple...
        assert isinstance(result, tuple)
        # containing 2 items (plugins, config)
        assert len(result) == 2
        # In the case of garbage in, we get garbage out; both lists are empty
        assert result == ([], [])

    # Valid Configuration
    result, config = ConfigBase.config_parse_text(
        """
    # A completely invalid token on json string (it gets ignored)
    # but the URL is still valid
    json://localhost?invalid-token=nodashes

    # A comment line over top of a URL
    mailto://userb:pass@gmail.com

    # Test a URL using it's native format; in this case Ryver
    https://apprise.ryver.com/application/webhook/ckhrjW8w672m6HG

    # Invalid URL as it's not associated with a plugin
    # or a native url
    https://not.a.native.url/

    # A line with mulitiple tag assignments to it
    taga,tagb=kde://

    # An include statement to Apprise API with trailing spaces:
    include http://localhost:8080/notify/apprise

    # A relative include statement (with trailing spaces)
    include apprise.cfg     """,
        asset=AppriseAsset(),
    )

    # We expect to parse 4 entries from the above
    assert isinstance(result, list)
    assert isinstance(config, list)
    assert len(result) == 4
    assert len(result[0].tags) == 0

    # Our last element will have 2 tags associated with it
    assert len(result[-1].tags) == 2
    assert "taga" in result[-1].tags
    assert "tagb" in result[-1].tags

    assert len(config) == 2
    assert "http://localhost:8080/notify/apprise" in config
    assert "apprise.cfg" in config

    # Here is a similar result set however this one has an invalid line
    # in it which invalidates the entire file
    result, config = ConfigBase.config_parse_text("""
    # A comment line over top of a URL
    mailto://userc:pass@gmail.com

    # A line with mulitiple tag assignments to it
    taga,tagb=windows://

    I am an invalid line that does not follow any of the Apprise file rules!
    """)

    # We expect to parse 0 entries from the above because the invalid line
    # invalidates the entire configuration file. This is for security reasons;
    # we don't want to point at files load content in them just because they
    # resemble an Apprise configuration.
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # More invalid data
    result, config = ConfigBase.config_parse_text("""
    # An invalid URL
    invalid://user:pass@gmail.com

    # A tag without a url
    taga=

    # A very poorly structured url
    sns://:@/

    # Just 1 token provided
    sns://T1JJ3T3L2/

    # Even with the above invalid entries, we can still
    # have valid include lines
    include file:///etc/apprise.cfg

    # An invalid include (nothing specified afterwards)
    include

    # An include of a config type we don't support
    include invalid://
    """)

    # We expect to parse 0 entries from the above
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Test case where a comment is on it's own line with nothing else
    result, config = ConfigBase.config_parse_text("#")
    # We expect to parse 0 entries from the above
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Verify our tagging works when multiple tags are provided
    result, config = ConfigBase.config_parse_text("""
    tag1, tag2, tag3=json://user:pass@localhost
    """)

    assert isinstance(result, list)
    assert len(result) == 1
    assert len(result[0].tags) == 3
    assert "tag1" in result[0].tags
    assert "tag2" in result[0].tags
    assert "tag3" in result[0].tags


def test_config_base_config_tag_groups_text():
    """
    API: ConfigBase.config_tag_groups_text object

    """

    # Valid Configuration
    result, config = ConfigBase.config_parse_text(
        """
    # Tag assignments
    groupA, groupB = tagB, tagC

    # groupB doubles down as it takes the entries initialized above
    # plus the added ones defined below
    groupB = tagA, tagB, tagD
    groupC = groupA, groupB, groupC, tagE

    # Tag that recursively looks to more tags
    groupD = groupC

    # Assigned ourselves
    groupX = groupX

    # Set up a recursive loop
    groupE = groupF
    groupF = groupE

    # Set up a larger recursive loop
    groupG = groupH
    groupH = groupI
    groupI = groupJ
    groupJ = groupK
    groupK = groupG

    # Bad assignments
    groupM = , , ,
     , ,   = , , ,

    # int's and floats are okay
    1 = 2
    a = 5

    # A comment line over top of a URL
    4, groupB = mailto://userb:pass@gmail.com

    # Tag Assignments
    tagA,groupB=json://localhost

    # More Tag Assignments
    tagC,groupB=xml://localhost

    # More Tag Assignments
    groupD=form://localhost

    """,
        asset=AppriseAsset(),
    )

    # We expect to parse 4 entries from the above
    assert isinstance(result, list)
    assert isinstance(config, list)
    assert len(result) == 4

    # Our first element is our group tags
    assert len(result[0].tags) == 2
    assert "groupB" in result[0].tags
    assert "4" in result[0].tags

    # No additional configuration is loaded
    assert len(config) == 0

    apobj = Apprise()
    assert apobj.add(result)
    # We match against 1 entry
    assert len(list(apobj.find("tagA"))) == 1
    assert len(list(apobj.find("tagB"))) == 0
    assert len(list(apobj.find("groupA"))) == 1
    assert len(list(apobj.find("groupB"))) == 3
    assert len(list(apobj.find("groupC"))) == 2
    assert len(list(apobj.find("groupD"))) == 3

    # Invalid Assignment
    result, config = ConfigBase.config_parse_text("""
    # Must have something to equal or it's a bad line
    group =

    # A tag Assignments that is never gotten to as the line
    # above is bad
    groupD=form://localhost
    """)

    # We expect to parse 0 entries from the above
    assert isinstance(result, list)
    assert isinstance(config, list)
    assert len(result) == 0
    assert len(config) == 0

    # Invalid Assignment
    result, config = ConfigBase.config_parse_text("""
    # Rundant assignment
    group = group

    # Our group assignment
    group=windows://

    """)

    # the redundant assignment does us no harm; but it doesn't grant us any
    # value either
    assert isinstance(result, list)
    assert len(result) == 1

    # Our first element is our group tags
    assert len(result[0].tags) == 1
    assert "group" in result[0].tags

    # There were no include entries defined
    assert len(config) == 0

    # More invalid data
    result, config = ConfigBase.config_parse_text("""
    # A tag without a url or group assignment
    taga=

    """)

    # We expect to parse 0 entries from the above
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    result, config = ConfigBase.config_parse_text("""
    # A tag without a url or group assignment
    taga= %%INVALID
    """)

    # We expect to parse 0 entries from the above
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0


def test_config_base_config_parse_text_with_url():
    """
    API: ConfigBase.config_parse_text object_with_url

    """
    # Here is a similar result set however this one has an invalid line
    # in it which invalidates the entire file
    result, config = ConfigBase.config_parse_text("""
    # Test a URL that has a URL as an argument
    json://user:pass@localhost?+arg=http://example.com?arg2=1&arg3=3
    """)

    # No tag is parsed, but our URL successfully parses as is

    assert isinstance(result, list)
    assert len(result) == 1
    assert len(result[0].tags) == 0

    # Verify our URL is correctly captured
    assert "%2Barg=http%3A%2F%2Fexample.com%3Farg2%3D1" in result[0].url()
    assert "json://user:pass@localhost/" in result[0].url()

    # There were no include entries defined
    assert len(config) == 0

    # Pass in our configuration again
    result, config = ConfigBase.config_parse_text(result[0].url())

    # Verify that our results repeat themselves
    assert isinstance(result, list)
    assert len(result) == 1
    assert len(result[0].tags) == 0
    assert "%2Barg=http%3A%2F%2Fexample.com%3Farg2%3D1" in result[0].url()
    assert "json://user:pass@localhost/" in result[0].url()

    assert len(config) == 0


def test_config_base_config_parse_yaml():
    """
    API: ConfigBase.config_parse_yaml object

    """

    # general reference used below
    asset = AppriseAsset()

    # Garbage Handling
    for garbage in (object(), None, "", 42):
        # A response is always correctly returned
        result = ConfigBase.config_parse_yaml(garbage)
        # response is a tuple...
        assert isinstance(result, tuple)
        # containing 2 items (plugins, config)
        assert len(result) == 2
        # In the case of garbage in, we get garbage out; both lists are empty
        assert result == ([], [])

    # Invalid Version
    result, config = ConfigBase.config_parse_yaml("version: 2a", asset=asset)

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Invalid Syntax (throws a ScannerError)
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

urls
""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Missing url token
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # No urls defined
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

urls:
""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Invalid url defined
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

# Invalid URL definition; yet the answer to life at the same time
urls: 43
""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Invalid url/schema
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

urls:
  - invalid://

""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Invalid url/schema
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

urls:
  - invalid://:
    - a: b

""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Invalid url/schema
    result, config = ConfigBase.config_parse_yaml(
        """
# Include entry with nothing associated with it
include:

urls:
  - just some free text that isn't valid:
    - a garbage entry to go with it

""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Invalid url/schema
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

urls:
  - not even a proper url

""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Invalid url/schema
    result, config = ConfigBase.config_parse_yaml(
        """
urls:
  # a very invalid sns entry
  - sns://T1JJ3T3L2/
  - sns://:@/:
    - invalid: test
  - sns://T1JJ3T3L2/:
    - invalid: test
    - _invalid: Token can not start with an underscore

  # some strangeness
  -
    -
      - test

""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # Valid Configuration
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

# Including by dict
include:
  # File includes
  - file:///absolute/path/
  - relative/path
  # Trailing colon shouldn't disrupt include
  - http://test.com:

  # invalid (numeric)
  - 4

  # some strangeness
  -
    -
      - test

#
# Define your notification urls:
#
urls:
  - pbul://o.gn5kj6nfhv736I7jC3cj3QLRiyhgl98b
  - mailto://test:password@gmail.com
  - https://apprise.ryver.com/application/webhook/ckhrjW8w672m6HG
  - https://not.a.native.url/

    # A completely invalid token on json string (it gets ignored)
    # but the URL is still valid
  - json://localhost?invalid-token=nodashes

""",
        asset=asset,
    )

    # We expect to parse 4 entries from the above
    # The Ryver one is in a native form and the 4th one is invalid
    assert isinstance(result, list)
    assert len(result) == 4
    assert len(result[0].tags) == 0

    # There were 3 include entries
    assert len(config) == 3
    assert "file:///absolute/path/" in config
    assert "relative/path" in config
    assert "http://test.com" in config

    # Valid Configuration
    result, config = ConfigBase.config_parse_yaml(
        """
# A single line include is supported
include: http://localhost:8080/notify/apprise

urls:
  # The following generates 1 service
  - json://localhost:
       tag: my-custom-tag, my-other-tag

  # The following also generates 1 service
  - json://localhost:
    - tag: my-custom-tag, my-other-tag

  # How to stack multiple entries (this generates 2):
  - mailto://user:123abc@yahoo.ca:
    - to: test@examle.com
    - to: test2@examle.com

      # This is an illegal entry; the schema can not be changed
      schema: json

  # accidently left a colon at the end of the url; no problem
  # we'll accept it
  - mailto://oscar:pass@gmail.com:

  # A Ryver URL (using Native format); still accepted
  - https://apprise.ryver.com/application/webhook/ckhrjW8w672m6HG:

  # An invalid URL with colon (ignored)
  - https://not.a.native.url/:

  # A telegram entry (returns a None in parse_url())
  - tgram://invalid

""",
        asset=asset,
    )

    # We expect to parse 6 entries from the above because the tgram:// entry
    # would have failed to be loaded
    assert isinstance(result, list)
    assert len(result) == 6
    assert len(result[0].tags) == 2

    # Our single line included
    assert len(config) == 1
    assert "http://localhost:8080/notify/apprise" in config

    # Global Tags
    result, config = ConfigBase.config_parse_yaml(
        """
# Global Tags stacked as a list
tag:
  - admin
  - devops

urls:
  - json://localhost
  - dbus://
""",
        asset=asset,
    )

    # We expect to parse 2 entries from the above
    assert isinstance(result, list)
    assert len(result) == 2

    # There were no include entries defined
    assert len(config) == 0

    # all entries will have our global tags defined in them
    for entry in result:
        assert "admin" in entry.tags
        assert "devops" in entry.tags

    # Global Tags
    result, config = ConfigBase.config_parse_yaml(
        """
# Global Tags
tag: admin, devops

urls:
  # The following tags will get added to the global set
  - json://localhost:
    - tag: string-tag, my-other-tag, text

  # Tags can be presented in this list format too:
  - dbus://:
    - tag:
      - list-tag
      - dbus
""",
        asset=asset,
    )

    # all entries will have our global tags defined in them
    for entry in result:
        assert "admin" in entry.tags
        assert "devops" in entry.tags

    # We expect to parse 2 entries from the above
    assert isinstance(result, list)
    assert len(result) == 2

    # json:// has 2 globals + 3 defined
    assert len(result[0].tags) == 5
    assert "text" in result[0].tags

    # json:// has 2 globals + 2 defined
    assert len(result[1].tags) == 4
    assert "list-tag" in result[1].tags

    # There were no include entries defined
    assert len(config) == 0

    # An invalid set of entries
    result, config = ConfigBase.config_parse_yaml(
        """
urls:
  # The following tags will get added to the global set
  - json://localhost:
    -
      -
        - entry
""",
        asset=asset,
    )

    # We expect to parse 0 entries from the above
    assert isinstance(result, list)
    assert len(result) == 0

    # There were no include entries defined
    assert len(config) == 0

    # An asset we'll manipulate; set some system flags
    asset = AppriseAsset(_uid="abc123", _recursion=1)

    # Global Tags
    result, config = ConfigBase.config_parse_yaml(
        """
# Test the creation of our apprise asset object
asset:
  app_id: AppriseTest
  app_desc: Apprise Test Notifications
  app_url: http://nuxref.com
  async_mode: no

  # System flags should never get set
  _uid: custom_id
  _recursion: 100

  # Support setting empty values
  image_url_mask:
  image_url_logo:

  image_path_mask: tmp/path

  # Timezone (supports tz keyword too)
  tz: America/Montreal

  # invalid entry
  theme:
    -
      -
        - entry

  # Now for some invalid entries
  invalid: entry
  __init__: can't be over-ridden
  nolists:
    - we don't support these entries
    - in the apprise object

urls:
  - json://localhost:
""",
        asset=asset,
    )

    # We expect to parse 1 entries from the above
    assert isinstance(result, list)
    assert len(result) == 1

    # There were no include entries defined
    assert len(config) == 0

    assert asset.app_id == "AppriseTest"
    assert asset.app_desc == "Apprise Test Notifications"
    assert asset.app_url == "http://nuxref.com"

    # Verify our system flags retain only the value they were initialized to
    assert asset._uid == "abc123"
    assert asset._recursion == 1

    # Boolean types stay boolean
    assert asset.async_mode is False

    # Our TimeZone
    assert isinstance(asset.tzinfo, tzinfo)
    assert asset.tzinfo.key == zoneinfo("America/Montreal").key

    # the theme was not updated and remains the same as it was
    assert asset.theme == AppriseAsset().theme

    # Empty string assignment
    assert isinstance(asset.image_url_mask, str)
    assert asset.image_url_mask == ""
    assert isinstance(asset.image_url_logo, str)
    assert asset.image_url_logo == ""

    # For on-lookers looking through this file; here is a perfectly formatted
    # YAML configuration file for your reference so you can see it without
    # all of the errors like the ones identified above
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed. Thus this is a
# completely optional field. It's a good idea to just add this line because it
# will help with future ambiguity (if it ever occurs).
version: 1

# Define an Asset object if you wish (Optional)
asset:
  app_id: AppriseTest
  app_desc: Apprise Test Notifications
  app_url: http://nuxref.com

  # An invalid timezone
  timezone: invalid

# Optionally define some global tags to associate with ALL of your
# urls below.
tag: admin, devops

# Define your URLs (Mandatory!)
urls:
  # Either on-line each entry like this:
  - json://localhost

  # Or add a colon to the end of the URL where you can optionally provide
  # over-ride entries.  One of the most likely entry to be used here
  # is the tag entry.  This gets extended to the global tag (if defined)
  # above
  - xml://localhost:
    - tag: customer

  # The more elements you specify under a URL the more times the URL will
  # get replicated and used. Hence this entry actually could be considered
  # 2 URLs being called with just the destination email address changed:
  - mailto://george:password@gmail.com:
     - to: jason@hotmail.com
     - to: fred@live.com

  # Again... to re-iterate, the above mailto:// would actually fire two (2)
  # separate emails each with a different destination address specified.
  # Be careful when defining your arguments and differentiating between
  # when to use the dash (-) and when not to.  Each time you do, you will
  # cause another instance to be created.

  # Defining more then 1 element to a muti-set is easy, it looks like this:
  - mailto://jackson:abc123@hotmail.com:
     - to: jeff@gmail.com
       tag: jeff, customer

     - to: chris@yahoo.com
       tag: chris, customer
""",
        asset=asset,
    )

    # okay, here is how we get our total based on the above (read top-down)
    # +1  json:// entry
    # +1  xml:// entry
    # +2  mailto:// entry to jason@hotmail.com and fred@live.com
    # +2  mailto:// entry to jeff@gmail.com and chris@yahoo.com
    # = 6
    assert len(result) == 6

    # all six entries will have our global tags defined in them
    for entry in result:
        assert "admin" in entry.tags
        assert "devops" in entry.tags

    # Entries can be directly accessed as they were added

    # our json:// had no additional tags added; so just the global ones
    # So just 2; admin and devops (these were already validated above in the
    # for loop
    assert len(result[0].tags) == 2

    # our xml:// object has 1 tag added (customer)
    assert len(result[1].tags) == 3
    assert "customer" in result[1].tags

    # You get the idea, here is just a direct mapping to the remaining entries
    # in the same order they appear above
    assert len(result[2].tags) == 2
    assert len(result[3].tags) == 2

    assert len(result[4].tags) == 4
    assert "customer" in result[4].tags
    assert "jeff" in result[4].tags

    assert len(result[5].tags) == 4
    assert "customer" in result[5].tags
    assert "chris" in result[5].tags

    # There were no include entries defined
    assert len(config) == 0

    # Valid Configuration (multi inline configuration entries)
    result, config = ConfigBase.config_parse_yaml(
        """
# A configuration file that contains 2 includes separated by a comma and/or
# space:
include: http://localhost:8080/notify/apprise, http://localhost/apprise/cfg

""",
        asset=asset,
    )

    # We will have loaded no results
    assert isinstance(result, list)
    assert len(result) == 0

    # But our two configuration files will be present:
    assert len(config) == 2
    assert "http://localhost:8080/notify/apprise" in config
    assert "http://localhost/apprise/cfg" in config

    # Valid Configuration (another way of specifying more then one include)
    result, config = ConfigBase.config_parse_yaml(
        """
# A configuration file that contains 4 includes on their own
# lines beneath the keyword `include`:
include:
   http://localhost:8080/notify/apprise
   http://localhost/apprise/cfg01
   http://localhost/apprise/cfg02
   http://localhost/apprise/cfg03

""",
        asset=asset,
    )

    # We will have loaded no results
    assert isinstance(result, list)
    assert len(result) == 0

    # But our 4 configuration files will be present:
    assert len(config) == 4
    assert "http://localhost:8080/notify/apprise" in config
    assert "http://localhost/apprise/cfg01" in config
    assert "http://localhost/apprise/cfg02" in config
    assert "http://localhost/apprise/cfg03" in config

    # Test a configuration with an invalid schema with options
    result, config = ConfigBase.config_parse_yaml(
        """
    urls:
      - invalid://:
          tag: 'invalid'
          :name: 'Testing2'
          :body: 'test body2'
          :title: 'test title2'
""",
        asset=asset,
    )

    # We will have loaded no results
    assert isinstance(result, list)
    assert len(result) == 0

    # Valid Configuration (we allow comma separated entries for
    # each defined bullet)
    result, config = ConfigBase.config_parse_yaml(
        """
# A configuration file that contains 4 includes on their own
# lines beneath the keyword `include`:
include:
   - http://localhost:8080/notify/apprise, http://localhost/apprise/cfg01
     http://localhost/apprise/cfg02
   - http://localhost/apprise/cfg03

""",
        asset=asset,
    )

    # We will have loaded no results
    assert isinstance(result, list)
    assert len(result) == 0

    # But our 4 configuration files will be present:
    assert len(config) == 4
    assert "http://localhost:8080/notify/apprise" in config
    assert "http://localhost/apprise/cfg01" in config
    assert "http://localhost/apprise/cfg02" in config
    assert "http://localhost/apprise/cfg03" in config


def test_config_base_config_parse_yaml_includes(
    requests_remote_config: Mock,
) -> None:
    """
    API: ConfigBase.config_parse_yaml_includes

    Verify that HTTP include entries are fetched via requests.get and that
    the remote config bodies are parsed into json:// and xml:// notifiers.
    """

    # general reference used below
    asset = AppriseAsset()

    # Initialize our apprise configuration
    ac = AppriseConfig(asset=asset, recursion=1)

    # Add our entry
    ac.add_config(
        cleandoc("""
        # Include our Apprise Configuration from 2 locations
        include:
           - http://localhost:8000/get/test-001
           - http://localhost:8000/get/test-002

        # no further URLs defined
    """)
    )

    # Force a fresh parse and get the loaded plugin
    servers = ac.servers()

    # the following will return
    assert len(servers) == 3

    # representation for NotifyBase subclasses.
    urls = {n.url() for n in servers}

    # The *exact* URL string may include extra params depending on defaults,
    # so we check using containment instead of strict equality.
    assert any(u.startswith("json://localhost") for u in urls)
    assert any(u.startswith("xml://localhost") for u in urls)
    assert any(u.startswith("form://localhost") for u in urls)


def test_yaml_vs_text_tagging():
    """
    API: ConfigBase YAML vs TEXT tagging
    """

    yaml_result, _ = ConfigBase.config_parse_yaml("""
    urls:
      - mailtos://lead2gold:yesqbrulvaelyxve@gmail.com:
         tag: mytag
    """)
    assert yaml_result

    text_result, _ = ConfigBase.config_parse_text("""
    mytag=mailtos://lead2gold:yesqbrulvaelyxve@gmail.com
    """)
    assert text_result

    # Now we compare our results and verify they are the same
    assert len(yaml_result) == len(text_result)
    assert isinstance(yaml_result[0], NotifyEmail)
    assert isinstance(text_result[0], NotifyEmail)
    assert "mytag" in text_result[0]
    assert "mytag" in yaml_result[0]


def test_config_base_config_tag_groups_yaml_01():
    """
    API: ConfigBase.config_tag_groups_yaml #1 object

    """

    # general reference used below
    asset = AppriseAsset()

    # Valid Configuration
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

groups:
  - group1: tagB, tagC, tagNotAssigned
  - group2:
      - tagA
      - tagC
  - group3:
      - tagD: optional comment
      - tagA: optional comment #2

  # No assignment
  - group4

  # No assignment type 2
  - group5:

  # Integer assignment
  - group6: 3
  - group6: 3, 4, 5, test
  - group6: 3.5, tagC

  # Recursion
  - groupA: groupB
  - groupB: groupA
  # And Again... (just because)
  - groupA: groupB
  - groupB: groupA

  # Self assignment
  - groupX: groupX

  # Set up a larger recursive loop
  - groupG: groupH
  - groupH: groupI, groupJ
  - groupI: groupJ, groupG
  - groupJ: groupK, groupH, groupI
  - groupK: groupG

  # No tags assigned
  - groupK: ",,  , ,"
  - " , ": ",, , ,"

  # Multi Assignments
  - groupL, groupM: tagD, tagA
  - 4, groupN:
     - tagD
     - tagE, TagA

  # Add one more tag to groupL making it different then GroupM by 1
  - groupL: tagB
#
# Define your notification urls:
#
urls:
  - form://localhost:
     - tag: tagA
  - mailto://test:password@gmail.com:
     - tag: tagB
  - xml://localhost:
     - tag: tagC
  - json://localhost:
     - tag: tagD, tagA

""",
        asset=asset,
    )

    # We expect to parse 4 entries from the above
    assert isinstance(result, list)
    assert isinstance(config, list)
    assert len(result) == 4

    # Our first element is our group tags
    assert len(result[0].tags) == 5
    assert "group2" in result[0].tags
    assert "group3" in result[0].tags
    assert "groupL" in result[0].tags
    assert "groupM" in result[0].tags
    assert "tagA" in result[0].tags

    # No additional configuration is loaded
    assert len(config) == 0

    apobj = Apprise()
    assert apobj.add(result)
    # We match against 1 entry
    assert len(list(apobj.find("tagA"))) == 2
    assert len(list(apobj.find("tagB"))) == 1
    assert len(list(apobj.find("tagC"))) == 1
    assert len(list(apobj.find("tagD"))) == 1
    assert len(list(apobj.find("group1"))) == 2
    assert len(list(apobj.find("group2"))) == 3
    assert len(list(apobj.find("group3"))) == 2
    assert len(list(apobj.find("group4"))) == 0
    assert len(list(apobj.find("group5"))) == 0
    # json:// -- group6 -> 4 -> TagA
    # xml://  -- group6 -> TagC
    assert len(list(apobj.find("group6"))) == 2
    assert len(list(apobj.find("4"))) == 1
    assert len(list(apobj.find("groupN"))) == 1


def test_config_base_config_tag_groups_yaml_02():
    """
    API: ConfigBase.config_tag_groups_yaml #2 object

    """

    # general reference used below
    asset = AppriseAsset()

    # Valid Configuration
    result, config = ConfigBase.config_parse_yaml(
        """
# if no version is specified then version 1 is presumed
version: 1

groups:
  group1: tagB, tagC, tagNotAssigned
  group2:
    - tagA
    - tagC
  group3:
    - tagD: optional comment
    - tagA: optional comment #2

  # No assignment type 2
  group5:

  # Integer assignment (since it's not a list, the last element prevails
  # and replaces the above); '4' does not get appended as it would in
  # the event this was a list instead
  group6: 3
  group6: 3, 4, 5, test
  group6: 3.5, tagC

  # Recursion
  groupA: groupB
  groupB: groupA
  # And Again... (just because)
  groupA: groupB
  groupB: groupA

  # Self assignment
  groupX: groupX

  # Set up a larger recursive loop
  groupG: groupH
  groupH: groupI, groupJ
  groupI: groupJ, groupG
  groupJ: groupK, groupH, groupI
  groupK: groupG

  # No tags assigned
  groupK: ",,  , ,"
  " , ": ",, , ,"

  # Multi Assignments
  groupL, groupM: tagD, tagA
  4, groupN:
   - tagD
   - tagE, TagA

  # Add one more tag to groupL making it different then GroupM by 1
  groupL: tagB
#
# Define your notification urls:
#
urls:
  - form://localhost:
     - tag: tagA
  - mailto://test:password@gmail.com:
     - tag: tagB
  - xml://localhost:
     - tag: tagC
  - json://localhost:
     - tag: tagD, tagA

""",
        asset=asset,
    )

    # We expect to parse 4 entries from the above
    assert isinstance(result, list)
    assert isinstance(config, list)
    assert len(result) == 4

    # Our first element is our group tags
    assert len(result[0].tags) == 5
    assert "group2" in result[0].tags
    assert "group3" in result[0].tags
    assert "groupL" in result[0].tags
    assert "groupM" in result[0].tags
    assert "tagA" in result[0].tags

    # No additional configuration is loaded
    assert len(config) == 0

    apobj = Apprise()
    assert apobj.add(result)
    # We match against 1 entry
    assert len(list(apobj.find("tagA"))) == 2
    assert len(list(apobj.find("tagB"))) == 1
    assert len(list(apobj.find("tagC"))) == 1
    assert len(list(apobj.find("tagD"))) == 1
    assert len(list(apobj.find("group1"))) == 2
    assert len(list(apobj.find("group2"))) == 3
    assert len(list(apobj.find("group3"))) == 2
    assert len(list(apobj.find("group4"))) == 0
    assert len(list(apobj.find("group5"))) == 0
    # NOT json:// -- group6 -> 4 -> TagA (not appended because dict storage)
    #                          ^
    #                          |
    #            See: test_config_base_config_tag_groups_yaml_01 (above)
    #                 dict storage (as this tests for) causes last entry to
    #                 prevail; previous assignments are lost
    #
    # xml://  -- group6 -> TagC
    assert len(list(apobj.find("group6"))) == 1
    assert len(list(apobj.find("4"))) == 1
    assert len(list(apobj.find("groupN"))) == 1
    assert len(list(apobj.find("groupK"))) == 0


def test_config_base_config_parse_yaml_globals():
    """
    API: ConfigBase.config_parse_yaml globals

    """

    # general reference used below
    asset = AppriseAsset()

    # Invalid Syntax (throws a ScannerError)
    results, config = ConfigBase.config_parse_yaml(
        cleandoc("""
    urls:
      - jsons://localhost1:
         - to: jeff@gmail.com
           tag: jeff, customer
           cto: 30
           rto: 30
           verify: no

      - jsons://localhost2?cto=30&rto=30&verify=no:
         - to: json@gmail.com
           tag: json, customer
    """),
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(results, list)

    # Our results loaded
    assert len(results) == 2
    assert len(config) == 0

    # Now verify that our global variables correctly initialized
    for entry in results:
        assert entry.verify_certificate is False
        assert entry.socket_read_timeout == 30
        assert entry.socket_connect_timeout == 30


# This test fails on CentOS 8.x so it was moved into it's own function
# so it could be bypassed. The ability to use lists in YAML files didn't
# appear to happen until later on; it's certainly not available in v3.12
# which was what shipped with CentOS v8 at the time.
@pytest.mark.skipif(
    int(yaml.__version__.split(".")[0]) <= 3,
    reason="requires pyaml v4.x or higher.",
)
def test_config_base_config_parse_yaml_list():
    """
    API: ConfigBase.config_parse_yaml list parsing

    """

    # general reference used below
    asset = AppriseAsset()

    # Invalid url/schema
    result, config = ConfigBase.config_parse_yaml(
        """
# no lists... just no
urls: [milk, pumpkin pie, eggs, juice]

# Including by list is okay
include: [file:///absolute/path/, relative/path, http://test.com]

""",
        asset=asset,
    )

    # Invalid data gets us an empty result set
    assert isinstance(result, list)
    assert len(result) == 0

    # There were 3 include entries
    assert len(config) == 3
    assert "file:///absolute/path/" in config
    assert "relative/path" in config
    assert "http://test.com" in config


def test_yaml_asset_timezone_and_asset_tokens(tmpdir):
    """
    Covers: valid tz, reserved keys, invalid key, bool coercion, None->"",
    invalid type for string, and %z formatting path used later by plugins.
    """
    cfg = tmpdir.join("asset-tz.yml")
    cfg.write(
        """
version: 1
asset:
  tz: "  america/toronto  "     # case-insensitive + whitespace cleanup
  _private: "ignored"           # reserved (starts with _)
  name_: "ignored"              # reserved (ends with _)
  not_a_field: "ignored"        # invalid asset key
  secure_logging: "yes"         # string -> bool via parse_bool
  app_id: null                  # None becomes empty string
  app_desc: [ "list" ]          # invalid type for string -> warning path
urls:
  - json://localhost
"""
    )

    ac = AppriseConfig(paths=str(cfg))
    # Force a fresh parse and get the loaded plugin
    servers = ac.servers()
    assert len(servers) == 1

    plugin = servers[0]
    asset = plugin.asset

    # tz was accepted and normalised
    # lower() is required since Mac and Window are not case sensitive and will
    # See output as it was passed in and not corrected per IANA
    assert getattr(asset.tzinfo, "key", None).lower() == "america/toronto"
    # boolean coercion applied
    assert asset.secure_logging is True
    # None -> ""
    assert asset.app_id == ""


def test_yaml_asset_timezone_invalid_and_precedence(tmpdir):
    """
    If 'timezone' is present but invalid, it takes precedence over 'tz'
    and MUST NOT set the asset to the 'tz' value. We assert that London
    was not applied. We deliberately avoid asserting the exact fallback,
    since environments may surface a system tz (datetime.timezone) that
    lacks a `.key` attribute.
    """
    cfg = tmpdir.join("asset-tz-invalid.yml")
    cfg.write(
        """
version: 1
asset:
  timezone: null                # invalid (will be seen as "None")
  tz: Europe/London             # would be valid, but 'timezone' wins
urls:
  - json://localhost
"""
    )

    base_asset = AppriseAsset(timezone="UTC")
    ac = AppriseConfig(paths=str(cfg))
    servers = ac.servers(asset=base_asset)
    assert len(servers) == 1

    tzinfo = servers[0].asset.tzinfo

    # The key assertion: 'tz' MUST NOT have been applied
    assert getattr(tzinfo, "key", "").lower() != "europe/london"

    # Sanity check that something sensible is set
    # Compare offsets at a fixed instant instead of object identity
    dt = datetime(2024, 1, 1, 12, 0, tzinfo=_tz.utc)
    assert tzinfo.utcoffset(dt) is not None


@pytest.mark.parametrize(
    "garbage_yaml",
    [
        "123",
        "3.1415",
        "true",
        "[UTC]",
        "{x: UTC}",
    ],
)
def test_yaml_asset_tz_garbage_types_only(tmpdir, garbage_yaml):
    """
    If only 'tz' is present and it is non-string, it is ignored.
    We assert it didn't become a real IANA zone (e.g., Europe/London),
    and that the tzinfo is usable.
    """
    cfg = tmpdir.join("asset-tz-garbage-only.yml")
    cfg.write(
        f"""
version: 1
asset:
  tz: {garbage_yaml}            # non-string -> warning path
urls:
  - json://localhost
"""
    )

    base_asset = AppriseAsset(timezone="UTC")
    ac = AppriseConfig(paths=str(cfg))
    servers = ac.servers(asset=base_asset)
    assert len(servers) == 1

    tzinfo = servers[0].asset.tzinfo

    # 1) Did not “accidentally” become a valid IANA from elsewhere.
    assert getattr(tzinfo, "key", "").lower() != "europe/london"

    # 2) tzinfo is usable (offset resolves at a fixed instant).
    dt = datetime(2024, 1, 1, 12, 0, tzinfo=_tz.utc)
    assert tzinfo.utcoffset(dt) is not None
    # also stable tzname resolution
    assert isinstance(tzinfo.tzname(dt), str)


def test_config_base_parse_yaml_file05_tags_alias_dict_form(tmpdir):
    """
    API: ConfigBase.parse_yaml_file (#5)

    Validate `tags` is accepted as an alias of `tag` in the dict form:
      - "schema://...":
           tags: a-tag

    Also implicitly verifies `tags` does not leak into plugin kwargs, because
    plugin instantiation would fail if an unexpected kwarg is passed through.
    """
    t = tmpdir.mkdir("tags-alias-dict-form").join("apprise.yml")
    t.write("""urls:
  - pover://nsisxnvnqixq39t0cw54pxieyvtdd9@2jevtmstfg5a7hfxndiybasttxxfku:
      tags: test1
  - pover://rg8ta87qngcrkc6t4qbykxktou0uug@tqs3i88xlufexwl8t4asglt4zp5wfn:
      tags: test2
  - pover://jcqgnlyq2oetea4qg3iunahj8d5ijm@evalvutkhc8ipmz2lcgc70wtsm0qpb:
      tags: test3
""")

    ac = AppriseConfig(paths=str(t))

    # The number of configuration files that exist
    assert len(ac) == 1

    # All entries should load
    assert len(ac.servers()) == 3

    a = Apprise()
    assert a.add(servers=ac) is True
    assert len(a) == 3

    # Verify tag matching works
    assert sum(1 for _ in a.find("no-match")) == 0
    assert sum(1 for _ in a.find("all")) == 3
    assert sum(1 for _ in a.find("test1")) == 1
    assert sum(1 for _ in a.find("test2")) == 1
    assert sum(1 for _ in a.find("test3")) == 1
    assert sum(1 for _ in a.find("test1, test3")) == 2


def test_config_base_parse_yaml_file06_tags_alias_list_form(tmpdir):
    """
    API: ConfigBase.parse_yaml_file (#6)

    Validate `tags` is accepted as an alias of `tag` in the list-of-dicts form:
      - "schema://...":
          - tags: a-tag
          - tags: another-tag

    Expected behaviour: expands into multiple entries (one per list item),
    and each is tagged appropriately.
    """
    t = tmpdir.mkdir("tags-alias-list-form").join("apprise.yml")
    t.write("""urls:
  - pover://nsisxnvnqixq39t0cw54pxieyvtdd9@2jevtmstfg5a7hfxndiybasttxxfku:
    - tags: test1
    - tags: test2
  - pover://jcqgnlyq2oetea4qg3iunahj8d5ijm@evalvutkhc8ipmz2lcgc70wtsm0qpb:
    - tags: test3
""")

    ac = AppriseConfig(paths=str(t))

    # The number of configuration files that exist
    assert len(ac) == 1

    # First URL expands to 2, second expands to 1
    assert len(ac.servers()) == 3

    a = Apprise()
    assert a.add(servers=ac) is True
    assert len(a) == 3

    # Verify tag matching works across expanded entries
    assert sum(1 for _ in a.find("test1")) == 1
    assert sum(1 for _ in a.find("test2")) == 1
    assert sum(1 for _ in a.find("test3")) == 1
    assert sum(1 for _ in a.find("test1, test2")) == 2


def test_config_base_parse_yaml_file07_tag_priority_over_tags(tmpdir):
    """
    API: ConfigBase.parse_yaml_file (#7)

    Validate priority: when both `tag` and `tags` are present, `tag` wins.

    This must remain true to preserve the original documented behaviour.
    """
    t = tmpdir.mkdir("tag-priority-over-tags").join("apprise.yml")
    t.write("""urls:
  - pover://nsisxnvnqixq39t0cw54pxieyvtdd9@2jevtmstfg5a7hfxndiybasttxxfku:
      tag: primary
      tags: secondary
""")

    ac = AppriseConfig(paths=str(t))

    # The number of configuration files that exist
    assert len(ac) == 1

    # Entry should load successfully
    assert len(ac.servers()) == 1

    a = Apprise()
    assert a.add(servers=ac) is True
    assert len(a) == 1

    # Tag priority check
    assert sum(1 for _ in a.find("primary")) == 1
    assert sum(1 for _ in a.find("secondary")) == 0



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import os
import sys

import helpers

logging.disable(logging.CRITICAL)

# Ensure we don't create .pyc files for these tests
sys.dont_write_bytecode = True


def test_environ_temporary_change():
    """helpers: environ() testing"""
    # This is a helper function; but it does enough that we want to verify
    # our usage of it works correctly; yes... we're testing a test

    e_key1 = "APPRISE_TEMP1"
    e_key2 = "APPRISE_TEMP2"
    e_key3 = "APPRISE_TEMP3"

    e_val1 = "ABCD"
    e_val2 = "DEFG"
    e_val3 = "HIJK"

    os.environ[e_key1] = e_val1
    os.environ[e_key2] = e_val2
    os.environ[e_key3] = e_val3

    # Ensure our environment variable stuck
    assert e_key1 in os.environ
    assert e_val1 in os.environ[e_key1]
    assert e_key2 in os.environ
    assert e_val2 in os.environ[e_key2]
    assert e_key3 in os.environ
    assert e_val3 in os.environ[e_key3]

    with helpers.environ(e_key1, e_key3):
        # Eliminates Environment Variable 1 and 3
        assert e_key1 not in os.environ
        assert e_key2 in os.environ
        assert e_val2 in os.environ[e_key2]
        assert e_key3 not in os.environ

    # after with is over, environment is restored to normal
    assert e_key1 in os.environ
    assert e_val1 in os.environ[e_key1]
    assert e_key2 in os.environ
    assert e_val2 in os.environ[e_key2]
    assert e_key3 in os.environ
    assert e_val3 in os.environ[e_key3]

    d_key = "APPRISE_NOT_SET"
    n_key = "APPRISE_NEW_KEY"
    n_val = "NEW_VAL"

    # Verify that our temporary variables (defined above) are not pre-existing
    # environemnt variables as we'll be setting them below
    assert n_key not in os.environ
    assert d_key not in os.environ

    # makes it easier to pass in the arguments
    updates = {
        e_key1: e_val3,
        e_key2: e_val1,
        n_key: n_val,
    }
    with helpers.environ(d_key, e_key3, **updates):
        # Attempt to eliminate an undefined key (silently ignored)
        # Eliminates Environment Variable 3
        # Environment Variable 1 takes on the value of Env 3
        # Environment Variable 2 takes on the value of Env 1
        # Set a brand new variable that previously didn't exist
        assert e_key1 in os.environ
        assert e_val3 in os.environ[e_key1]
        assert e_key2 in os.environ
        assert e_val1 in os.environ[e_key2]
        assert e_key3 not in os.environ

        # Can't delete a variable that doesn't exist; so we're in the same
        # state here.
        assert d_key not in os.environ

        # Our temporary variables will be found now
        assert n_key in os.environ
        assert n_val in os.environ[n_key]

    # after with is over, environment is restored to normal
    assert e_key1 in os.environ
    assert e_val1 in os.environ[e_key1]
    assert e_key2 in os.environ
    assert e_val2 in os.environ[e_key2]
    assert e_key3 in os.environ
    assert e_val3 in os.environ[e_key3]

    # Even our temporary variables are now missing
    assert n_key not in os.environ
    assert d_key not in os.environ



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import time
from typing import ClassVar
from unittest import mock

import pytest
import requests

from apprise import NotificationManager
from apprise.common import ConfigFormat
from apprise.config.http import ConfigHTTP
from apprise.plugins import NotifyBase

logging.disable(logging.CRITICAL)

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()

# Some exception handling we'll use
REQUEST_EXCEPTIONS = (
    requests.ConnectionError(0, "requests.ConnectionError() not handled"),
    requests.RequestException(0, "requests.RequestException() not handled"),
    requests.HTTPError(0, "requests.HTTPError() not handled"),
    requests.ReadTimeout(0, "requests.ReadTimeout() not handled"),
    requests.TooManyRedirects(0, "requests.TooManyRedirects() not handled"),
)


@mock.patch("requests.post")
def test_config_http(mock_post):
    """
    API: ConfigHTTP() object

    """

    # Define our good:// url
    class GoodNotification(NotifyBase):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def notify(self, *args, **kwargs):
            # Pretend everything is okay
            return True

        def url(self, **kwargs):
            # Support url() function
            return ""

    # Store our good notification in our schema map
    N_MGR["good"] = GoodNotification

    # Our default content
    default_content = """taga,tagb=good://server01"""

    class DummyResponse:
        """A dummy response used to manage our object."""

        status_code = requests.codes.ok
        headers: ClassVar[dict[str, str]] = {
            "Content-Length": str(len(default_content)),
            "Content-Type": "text/plain",
        }

        text = default_content

        # Pointer to file
        ptr = None

        def close(self):
            return

        def raise_for_status(self):
            return

        def __enter__(self):
            return self

        def __exit__(self, *args, **kwargs):
            return

    # Prepare Mock
    dummy_response = DummyResponse()
    mock_post.return_value = dummy_response

    assert ConfigHTTP.parse_url("garbage://") is None

    results = ConfigHTTP.parse_url("http://user:pass@localhost?+key=value")
    assert isinstance(results, dict)
    ch = ConfigHTTP(**results)
    assert isinstance(ch.url(), str) is True
    assert isinstance(ch.read(), str) is True

    # one entry added
    assert len(ch) == 1

    results = ConfigHTTP.parse_url("http://localhost:8080/path/")
    assert isinstance(results, dict)
    ch = ConfigHTTP(**results)
    assert isinstance(ch.url(), str) is True
    assert isinstance(ch.read(), str) is True

    # one entry added
    assert len(ch) == 1

    # Clear all our mock counters
    mock_post.reset_mock()

    # Cache Handling; cache each request for 30 seconds
    results = ConfigHTTP.parse_url("http://localhost:8080/path/?cache=30")
    assert mock_post.call_count == 0
    assert isinstance(ch.url(), str) is True

    assert isinstance(results, dict)
    ch = ConfigHTTP(**results)
    assert mock_post.call_count == 0

    assert isinstance(ch.url(), str) is True
    assert mock_post.call_count == 0

    assert isinstance(ch.read(), str) is True
    assert mock_post.call_count == 1

    # Clear all our mock counters
    mock_post.reset_mock()

    # Behind the scenes we haven't actually made a fetch yet. We can consider
    # our content expired at this point
    assert ch.expired() is True

    # Test using boolean check; this will force a remote fetch
    assert ch

    # Now a call was made
    assert mock_post.call_count == 1
    mock_post.reset_mock()

    # Our content hasn't expired yet (it's good for 30 seconds)
    assert ch.expired() is False
    assert len(ch) == 1
    assert mock_post.call_count == 0

    # Test using boolean check; we will re-use our cache and not
    # make another remote request
    mock_post.reset_mock()
    assert ch
    assert len(ch.servers()) == 1
    assert len(ch) == 1

    # No remote post has been made
    assert mock_post.call_count == 0

    with mock.patch("time.time", return_value=time.time() + 10):
        # even with 10 seconds elapsed, no fetch will be made
        assert ch.expired() is False
        assert ch
        assert len(ch.servers()) == 1
        assert len(ch) == 1

    # No remote post has been made
    assert mock_post.call_count == 0

    with mock.patch("time.time", return_value=time.time() + 31):
        # but 30+ seconds from now is considered expired
        assert ch.expired() is True
        assert ch
        assert len(ch.servers()) == 1
        assert len(ch) == 1

    # Our content would have been renewed with a single new fetch
    assert mock_post.call_count == 1

    # one entry added
    assert len(ch) == 1

    # Invalid cache
    results = ConfigHTTP.parse_url("http://localhost:8080/path/?cache=False")
    assert isinstance(results, dict)
    assert isinstance(ch.url(), str) is True

    results = ConfigHTTP.parse_url("http://localhost:8080/path/?cache=-10")
    assert isinstance(results, dict)
    with pytest.raises(TypeError):
        ch = ConfigHTTP(**results)

    results = ConfigHTTP.parse_url("http://user@localhost?format=text")
    assert isinstance(results, dict)
    ch = ConfigHTTP(**results)
    assert isinstance(ch.url(), str) is True
    assert isinstance(ch.read(), str) is True

    # one entry added
    assert len(ch) == 1

    results = ConfigHTTP.parse_url("https://localhost")
    assert isinstance(results, dict)
    ch = ConfigHTTP(**results)
    assert isinstance(ch.url(), str) is True
    assert isinstance(ch.read(), str) is True

    # one entry added
    assert len(ch) == 1

    # Testing of pop
    ch = ConfigHTTP(**results)

    ref = ch[0]
    assert isinstance(ref, NotifyBase) is True

    ref_popped = ch.pop(0)
    assert isinstance(ref_popped, NotifyBase) is True

    assert ref == ref_popped

    assert len(ch) == 0

    # reference to calls on initial reference
    ch = ConfigHTTP(**results)
    assert isinstance(ch.pop(0), NotifyBase) is True

    ch = ConfigHTTP(**results)
    assert isinstance(ch[0], NotifyBase) is True
    # Second reference actually uses cache
    assert isinstance(ch[0], NotifyBase) is True

    ch = ConfigHTTP(**results)
    # Itereator creation (nothing needed to assert here)
    iter(ch)
    # Second reference actually uses cache
    iter(ch)

    # Test a buffer size limit reach
    ch.max_buffer_size = len(dummy_response.text)
    assert isinstance(ch.read(), str) is True

    # Test YAML detection
    yaml_supported_types = (
        "text/yaml",
        "text/x-yaml",
        "application/yaml",
        "application/x-yaml",
    )

    for st in yaml_supported_types:
        dummy_response.headers["Content-Type"] = st
        ch.default_config_format = None
        assert isinstance(ch.read(), str) is True
        # Set to YAML
        assert ch.default_config_format == ConfigFormat.YAML

    # Test TEXT detection
    text_supported_types = ("text/plain", "text/html")

    for st in text_supported_types:
        dummy_response.headers["Content-Type"] = st
        ch.default_config_format = None
        assert isinstance(ch.read(), str) is True
        # Set to TEXT
        assert ch.default_config_format == ConfigFormat.TEXT

    # The type is never adjusted to mime types we don't understand
    ukwn_supported_types = ("text/css", "application/zip")

    for st in ukwn_supported_types:
        dummy_response.headers["Content-Type"] = st
        ch.default_config_format = None
        assert isinstance(ch.read(), str) is True
        # Remains unchanged
        assert ch.default_config_format is None

    # When the entry is missing; we handle this too
    del dummy_response.headers["Content-Type"]
    ch.default_config_format = None
    assert isinstance(ch.read(), str) is True
    # Remains unchanged
    assert ch.default_config_format is None

    # Restore our content type object for lower tests
    dummy_response.headers["Content-Type"] = "text/plain"

    # Take a snapshot
    max_buffer_size = ch.max_buffer_size

    ch.max_buffer_size = len(dummy_response.text) - 1
    assert ch.read() is None

    # Restore buffer size count
    ch.max_buffer_size = max_buffer_size

    # Test erroneous Content-Length
    # Our content is still within the limits, so we're okay
    dummy_response.headers["Content-Length"] = "garbage"

    assert isinstance(ch.read(), str) is True

    dummy_response.headers["Content-Length"] = "None"
    # Our content is still within the limits, so we're okay
    assert isinstance(ch.read(), str) is True

    # Handle cases where the content length is exactly at our limit
    dummy_response.text = "a" * ch.max_buffer_size
    # This is acceptable
    assert isinstance(ch.read(), str) is True

    # If we are over our limit though..
    dummy_response.text = "b" * (ch.max_buffer_size + 1)
    assert ch.read() is None

    # Test an invalid return code
    dummy_response.status_code = 400
    assert ch.read() is None
    ch.max_error_buffer_size = 0
    assert ch.read() is None

    # Exception handling
    for exception in REQUEST_EXCEPTIONS:
        mock_post.side_effect = exception
        assert ch.read() is None

    # Restore buffer size count
    ch.max_buffer_size = max_buffer_size
